from ktdk.runtime.tags import TagsEvaluator


def get_evaluator(expr):
    resolver = TagsEvaluator(expr)
    resolver.register_tags('hello', 'world', 'naostro', 'nanecisto')
    return resolver


def test_tags_evaluator_empty():
    resolver = get_evaluator("")
    assert resolver.evaluate()
    assert resolver.evaluate('hello', 'world')
    assert resolver.evaluate("hello")


def test_tags_evaluator_simple_only_naostro():
    resolver = get_evaluator("naostro")
    assert resolver.evaluate("naostro")
    assert resolver.evaluate("naostro", "nanecisto")
    assert not resolver.evaluate("nanecisto")
    assert not resolver.evaluate()


def test_tags_evaluator_simple_or():
    resolver = get_evaluator("naostro or nanecisto")
    assert resolver.evaluate("naostro", "nanecisto")
    assert resolver.evaluate("naostro", "nanecisto", 'world')
    assert resolver.evaluate("naostro")
    assert resolver.evaluate("nanecisto")
    assert not resolver.evaluate()
    assert not resolver.evaluate("hello", "world")


def test_tags_evaluator_simple_and():
    resolver = get_evaluator("naostro and nanecisto")
    assert not resolver.evaluate()
    assert not resolver.evaluate("nanecisto")
    assert not resolver.evaluate("naostro")
    assert not resolver.evaluate("hello", "world")
    assert resolver.evaluate("naostro", "nanecisto")


def test_tags_evaluator_simple_not():
    resolver = get_evaluator("not naostro")
    assert resolver.evaluate()
    assert resolver.evaluate("hello", "world")
    assert resolver.evaluate("nanecisto")
    assert not resolver.evaluate("naostro")
    assert not resolver.evaluate("naostro", "nanecisto")


def test_tags_evaluator_using_non_existing_tag_in_test_tags():
    resolver = get_evaluator("naostro")
    assert resolver.evaluate("naostro")
    assert not resolver.evaluate("foo")
    assert resolver.evaluate("naostro", "bar")


def test_tags_evaluator_using_non_existing_tag_in_the_expression():
    resolver = get_evaluator("foo")
    assert not resolver.evaluate()
    assert not resolver.evaluate("foo")
    assert not resolver.evaluate("naostro", "bar")



from json import JSONEncoder

from ktdk.utils.basic import BasicObject


class Result(BasicObject, JSONEncoder):
    FAIL = "FAIL"
    ERROR = "ERROR"
    PASS = "PASS"
    SKIP = "SKIP"
    NONE = "NONE"

    def __init__(self, state: str):
        super().__init__()
        self._state = state

    @property
    def state(self):
        return self._state

    @property
    def passed(self):
        return self.state == Result.PASS

    @property
    def ok(self):
        return self.passed or self.state == Result.SKIP

    @property
    def nok(self):
        return not self.ok

    @property
    def errored(self):
        return self.state == Result.ERROR

    @property
    def failed(self):
        return self.state == Result.FAIL

    @property
    def skipped(self):
        return self.state == Result.SKIP

    @property
    def reason(self):
        return f'{self.state}'

    def to_dict(self):
        return {
            "state": self.state,
            "ok": self.ok
            }


PASS = Result(Result.PASS)
ERROR = Result(Result.ERROR)
FAIL = Result(Result.FAIL)
SKIP = Result(Result.SKIP)
NONE = Result(Result.NONE)



import logging

from ktdk import Task, Test
from ktdk.asserts import checks, matchers
from ktdk.asserts.checks import ReturnCodeMatchesCheck, StdErrMatchesCheck, \
    ValgrindPassedCheck
from ktdk.asserts.matchers import Equals, IsEmpty
from ktdk.tasks.command import Command
from ktdk.tasks.cpp.valgrind import ValgrindCommand
from ktdk.tasks.raw.executable import ExecutableTask
from ktdk.tasks.xunit import junit
from ktdk.utils import flatters, naming

log = logging.getLogger(__name__)


class CatchRunTestsOneByOneTask(Task):
    def __init__(self, executable=None, executor=Command, test_args=None, applied_check=None,
                 **kwargs):
        super().__init__(**kwargs)
        self._executable = executable
        self.executor = executor
        self.applied_check = applied_check
        self._test_args = test_args or []

    @property
    def executable(self):
        return self._executable

    @property
    def command_path(self):
        return self.context.config['exec'][self.executable]

    # https://github.com/catchorg/Catch2/blob/master/docs/command-line.md
    def get_the_tests_list(self):
        result = self.run_only_command(self.command_path, '--list-test-names-only',
                                       executor=Command)
        stdout = result.stdout.content
        return stdout.splitlines()

    def run_one_test_by_name(self, test_name):
        test = Test(name=test_name, desc=test_name)
        params = [*self._test_args, '-r', 'junit', '-n', naming.slugify(test_name), test_name]
        exec_task = self.run_command(self.executable, *params)
        test.add_task(exec_task)
        if self.test is not None:
            self.test.add_test(test)  # Add the newly created test to the test
        return exec_task

    def run_command(self, binary, *params, executor=None):
        executor = executor or self.executor
        task = ExecutableTask(binary, *params, executor=executor)
        return task

    def _run(self, *args, **kwargs):
        tests = self.get_the_tests_list()
        for test_name in tests:
            log.debug(f"[CATCH_RUN] Catch run the test: {test_name}")
            task = self.run_one_test_by_name(test_name=test_name)
            self.process_test_run(task)
            self.check_test_result(task, task.test)

    def run_only_command(self, binary, *params, executor=Command):
        cmd = executor(binary, args=params)
        cmd.set_task(self)
        return cmd.invoke()

    def check_test_result(self, task: Task, test=None):
        test = self.test

        def __default_check(task, test, executor=None):
            task.check_that(ReturnCodeMatchesCheck(matcher=Equals(0)))
            task.check_that(StdErrMatchesCheck(matcher=IsEmpty()))
            if executor == ValgrindCommand:
                task.check_that(ValgrindPassedCheck())

        if self.applied_check is not None:
            return self.applied_check(task, test, self.executor)
        else:
            return __default_check(task=task, test=test, executor=self.executor)

    def process_test_run(self, task):
        log.debug(f"[CATCH] Process and run: {task.test.namespace}")
        task.add_task(_ProcessJUnitTestRunTask())


class _ProcessJUnitTestRunTask(Task):
    def _run(self, *args, **kwargs):
        log.debug(f"[PROC_JUNIT] JUNIT process: {self.test.namespace} ")
        test = self.test
        exec_result = self.context.config['exec_result']
        stdout_path = exec_result.stdout.path
        if stdout_path:
            self.add_task(junit.JUnitParseTask(junit_file=stdout_path))
        return test


class CatchCheckAndComputePoint(Task):
    def __init__(self, max_points=1.0, **kwargs):
        super().__init__(**kwargs)
        self._max_points = max_points

    def _run(self, *args, **kwargs):
        tests = flatters.flatten_tests(self.test, include_self=False)
        num_tests = len(tests)
        for test in tests:
            test.points = self._max_points / num_tests
            test.check_that(checks.XUnitSuiteErrorsCheck(matcher=matchers.Equals(0)))
            test.check_that(checks.XUnitSuiteFailsCheck(matcher=matchers.Equals(0)))
            test.check_that(checks.ValgrindPassedCheck())


class CatchCheckCasesAndComputePoint(Task):
    def __init__(self, max_points=1.0, **kwargs):
        super().__init__(**kwargs)
        self._max_points = max_points

    def _run(self, *args, **kwargs):
        tests = flatters.flatten_tests(self.test, include_self=False)
        tests = [test for test in tests if 'junit_case' in test.tags]
        num_tests = len(tests)
        for test in tests:
            test.points = self._max_points / num_tests
            test.check_that(checks.XUnitCaseCheck(matcher=matchers.Equals(0)))
            test.check_that(checks.ValgrindPassedCheck())



from ktdk import KTDK, Task, Test
from ktdk.tasks import fs


class Scenario(Task):
    @property
    def ktdk(self) -> KTDK:
        return KTDK.instance

    @property
    def root_test(self) -> Test:
        return self.test

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class FullScenario(Scenario):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def ft(self):
        ft = fs.FileTasks()
        self.root_test.add_task(ft)
        return ft

    def file_tasks(self):
        pass

    def compile_tasks(self):
        pass

    def run_tasks(self):
        pass

    def evaluate_tasks(self):
        pass

    def _run(self, *args, **kwargs):
        self.file_tasks()
        self.compile_tasks()
        self.run_tasks()
        self.evaluate_tasks()



import string
from random import choice

from ktdk import Context
from ktdk.core.tasks import Task
from ktdk.core.tests import Test

ALPHABET = string.ascii_letters + string.digits


def get_test_context(**kwargs):
    params = {**kwargs}
    if params.get('suite_config'):
        params['suite_config']['devel'] = True
    return Context(**params)


def random_string(length=10, alpha=ALPHABET):
    return ''.join(choice(alpha) for i in range(length))


class TestUtils(object):
    EMPTY = {
        'name': 'test',
        'desc': None,
        'tags': set(),
        'tests': [],
        'before': [],
        'after': [],
    }

    @staticmethod
    def get_params(**params):
        defaults = {
            'name': random_string(alpha=string.ascii_lowercase),
            'desc': random_string(),
            'tags': ['naostro', 'generic', 'stylecheck'],
        }
        return {**TestUtils.EMPTY, **defaults, **params}

    @staticmethod
    def create_test(**params):
        params = TestUtils.get_params(**params)
        return Test(**params)

    @staticmethod
    def construct_test(**params):
        params = TestUtils.get_params(**params)
        return Test(**params)


class TaskUtils:
    EMPTY = dict(name='task', desc=None)

    @staticmethod
    def get_params(**params):
        defaults = {
            'name': random_string(alpha=string.ascii_lowercase),
            'desc': random_string(),
        }
        return {**TaskUtils.EMPTY, **defaults, **params}

    @staticmethod
    def create_task(**params):
        params = TaskUtils.get_params(**params)
        return Task(**params)



import logging

from ktdk.asserts.matchers import IsNotEmpty
from ktdk.core.tasks import Task
from ktdk.tasks.build_task import BuildTask
from ktdk.tasks.fs.tasks import FindFiles

log = logging.getLogger(__name__)


class RawCompilerTask(Task):
    class InternalCompilerTask(BuildTask):
        def __init__(self, executable, use_default_options=True, **kwargs):
            super().__init__(**kwargs)
            self.executable = executable
            self._use_default_options = use_default_options

        @property
        def compiler_options(self):
            return self.context.config['cpp_compiler_flags']

        def process_result(self, exec_path):
            self.context.config.add_test('exec', {self.executable: exec_path})

        def run_compile(self):
            found_files = self.context.config['found_files']
            self.asserts.require(found_files, matcher=IsNotEmpty("No input files to compile!"))
            str_files = [str(file_to_str) for file_to_str in found_files]
            build_dir = self.get_build_dir()
            exec_path = build_dir / self.executable
            if self.compiler_options and self._use_default_options:
                self.args.extend(self.compiler_options)
            self.args += ['-o', str(exec_path)]
            self.args += str_files
            self.execute()
            self.process_result(exec_path)

        def get_build_dir(self):
            return self.context.paths.workspace_path('build', create=True)

        def _run(self, *args, **kwargs):
            log.debug('[BLD] Compile')
            self.move_to_dir(self.target)
            self.run_compile()

    def __init__(self, executable, files, compiler=None, use_default_options=True, **kwargs):
        super().__init__(**kwargs)
        self.add_tags('compile')
        self.files = files
        self._compiler = compiler
        self.executable = executable
        self._use_default_options = use_default_options

    @property
    def compiler(self):
        if self._compiler is None:
            self._compiler = self.context.config.get('cpp_compiler')
        return self._compiler

    def _run(self, *args, **kwargs):
        find_files = FindFiles(pattern=self.files, from_dir=self.context.paths.workspace)
        find_files.add_task(RawCompilerTask.InternalCompilerTask(
            command=self.compiler,
            executable=self.executable,
            use_default_options=self._use_default_options
        ))
        self.add_task(find_files)


class CCompilerTask(RawCompilerTask):
    @property
    def compiler_options(self):
        return self.context.config['c_compiler_flags']

    @property
    def compiler(self):
        if self._compiler is None:
            self._compiler = self.context.config.get('c_compiler')
        return self._compiler


class CPPCompilerTask(RawCompilerTask):
    @property
    def compiler(self):
        if self._compiler is None:
            self._compiler = self.context.config.get('cpp_compiler')
        return self._compiler

    @property
    def compiler_options(self):
        return self.context.config['cpp_compiler_flags']



import junitparser

from ktdk.asserts.checks.general import AbstractMatchesTask


class AbstractXUnitSuiteCheck(AbstractMatchesTask):
    @property
    def suite(self) -> junitparser.TestSuite:
        return self.context.config['xunit_suite']


class AbstractXUnitCaseCheck(AbstractMatchesTask):
    @property
    def case(self) -> junitparser.TestCase:
        return self.context.config['xunit_case']

class XUnitSuiteFailsCheck(AbstractXUnitSuiteCheck):
    def _run(self, *args, **kwargs):
        self.asserts.check(self.suite.failures, matcher=self.matcher)


class XUnitSuiteErrorsCheck(AbstractXUnitSuiteCheck):
    def _run(self, *args, **kwargs):
        self.asserts.check(self.suite.errors, matcher=self.matcher)


class XUnitSuiteNokCheck(AbstractXUnitSuiteCheck):
    def _run(self, *args, **kwargs):
        noks = self.suite.errors + self.suite.failures
        self.asserts.check(noks, matcher=self.matcher)


class XUnitCaseCheck(AbstractXUnitCaseCheck):
    def _run(self, *args, **kwargs):
        self.asserts.check(self.case.result, matcher=self.matcher)




from typing import List

import ktdk.tasks.cpp.checkstyle as cpp_style
from ktdk import Test
from ktdk.asserts.checks import ClangTidyResultMatchesCheck, StyleResultMatchesCheck
from ktdk.asserts.matchers import IsEmpty, StdoutIsEmpty
from ktdk.scenarios import scenario


class StyleCheckScenario(scenario.Scenario):
    def __init__(self, files=None, headers=None, files_glob=None,
                 checkstyle_points=0, tidy_points=0, style=None,
                 clang_tidy_klass=cpp_style.CPPClangTidyCheckStyle, **kwargs):
        super().__init__(**kwargs)
        self._files = files
        self._headers = headers
        self._files_glob = files_glob
        self._checkstyle_points = checkstyle_points
        self._tidy_points = tidy_points
        self._clang_tidy_klass = clang_tidy_klass
        self._format_style = style

    @property
    def files(self) -> List:
        if self._files:
            return self._files
        else:
            return list(self.context.paths.workspace.glob(self._files_glob))

    def run_clang_tidy(self, files: list = None):
        files = files or self.files
        tidy_test = Test(name="clang_tidy", desc="Clang Tidy test for given files",
                         points=self._tidy_points)
        tidy_task = self._clang_tidy_klass(files=files, headers=self._headers)
        tidy_task.check_that(ClangTidyResultMatchesCheck(matcher=StdoutIsEmpty()))
        tidy_test.add_task(tidy_task)
        self.test.add_test(tidy_test)

    def run_clang_format_test(self, file, points: float = 0):
        format_test = Test(name="clang_format", desc=f"Clang format tests for {file.name}",
                           points=points)
        self.test.add_test(format_test)
        format_task = cpp_style.ClangFormatTask(file=file, style=self._format_style)
        format_task.check_that(StyleResultMatchesCheck(matcher=IsEmpty()))
        format_test.add_task(format_task)

    def _run(self, *args, **kwargs):
        self.run_clang_tidy()
        files = self.files
        files_count = len(files)
        part = self._checkstyle_points / files_count
        for file in files:
            self.run_clang_format_test(file, points=part)



import os
import shutil
from pathlib import Path

import pytest

from ktdk import KTDK, Test
from ktdk.asserts import matchers, checks
from ktdk.tasks.cpp import CMakeBuildTask, CatchCheckCasesAndComputePoint, ValgrindCommand, catch
from ktdk.tasks.fs import CopyFiles, ExistFiles, MakeDir
from ktdk.tasks.fs.tools import FileTasks
from tests.paths import TEST_RESOURCES_BASE


def create_cmake(src, *names):
    content = """
    cmake_minimum_required (VERSION 3.0)
    project (kontr_tests)
    
    add_compile_options(-std=c++11)
    
    """
    src = Path(src)
    for name in names:
        content += f"add_executable({name} {name}.cpp)\n"
    cmake_file = src.joinpath('CMakeLists.txt')
    cmake_file.write_text(content)


@pytest.fixture
def workspace_dir(tmpdir):
    workspace = tmpdir.mkdir('workspace')
    return workspace


@pytest.fixture
def test_files_dir(tmpdir):
    test_files = tmpdir.mkdir('test_files')
    return test_files


@pytest.fixture
def submission_dir(tmpdir):
    submission = tmpdir.mkdir('submission')
    return submission


@pytest.fixture()
def prepared_sources(submission_dir):
    path = os.path.join(submission_dir, 'src')
    shutil.copytree(str(TEST_RESOURCES_BASE / 'catch'), path)
    create_cmake(path, 'tests')
    return submission_dir


@pytest.fixture
def cmake():
    cmake = CMakeBuildTask(source='src', cwd='src')
    cmake.require_that(checks.TaskResultCheck(matcher=matchers.ResultPassed()))
    return cmake


@pytest.fixture()
def file_tasks():
    ft = FileTasks()
    ft.workspace('src').add_task(MakeDir())
    ft.submission('src').require_that(ExistFiles('tests.cpp'))

    ft.submission('src').add_task(CopyFiles('*.*'))
    ft.submission('src').add_task(CopyFiles('CMakeLists.txt'))
    ft.workspace('src').check_that(ExistFiles("CMakeLists.txt"))
    ft.workspace('src').check_that(ExistFiles("tests.cpp"))
    return ft


@pytest.fixture()
def catch_test():
    catch_test = Test(name="catch_test", desc="Super catch test")
    catch_task = catch.CatchRunTestsOneByOneTask(executable='tests', executor=ValgrindCommand)
    catch_test.add_task(catch_task)
    catch_test.check_that(CatchCheckCasesAndComputePoint(max_points=3), after_tasks=True)
    return catch_test


@pytest.fixture()
def naostro(catch_test):
    naostro = Test(name="naostro", desc="Test nanecisto", tags=['naostro'])
    naostro.add_test(catch_test)
    return naostro


@pytest.fixture()
def ktdk(workspace_dir, test_files_dir, submission_dir, cmake, naostro, file_tasks):
    ktdk = KTDK(
        test_files=test_files_dir,
        submission=submission_dir,
        workspace=workspace_dir,
        devel=True
    )
    ktdk.suite.add_task(file_tasks)
    ktdk.suite.add_task(cmake)
    ktdk.suite.add_test(naostro)
    return ktdk


def test_catch_run(prepared_sources, ktdk):
    ktdk.invoke()
    assert ktdk.suite.result.effective.passed
    assert round(ktdk.suite.result.effective_points, 3) == 3



from typing import Optional

from ktdk.asserts.checks.general import AbstractMatchesTask


class AbstractStyleResultMatchesTask(AbstractMatchesTask):
    @property
    def style_result(self) -> Optional[str]:
        if not self.context:
            return None
        return self.context.config['tidy_result']


class StyleResultMatchesCheck(AbstractStyleResultMatchesTask):
    def _run(self, *args, **kwargs):
        self.asserts.check(self.style_result, self.matcher)


class ClangTidyResultMatchesCheck(AbstractMatchesTask):
    @property
    def tidy_result(self) -> Optional[str]:
        if not self.context:
            return None
        return self.context.config['tidy_result']

    def _run(self, *args, **kwargs):
        self.asserts.check(self.tidy_result, self.matcher)



from ktdk.core.tasks import Task


class AbstractMatchesTask(Task):
    def __init__(self, matcher=None, **kwargs):
        super().__init__(**kwargs)
        self.matcher = matcher
        self._asserted_object = None

    def _run(self, *args, **kwargs):
        self.asserts.check(self._asserted_object, self.matcher)


class TaskResultCheck(AbstractMatchesTask):
    def _run(self, *args, **kwargs):
        self.asserts.check(self.parent.result.current, self.matcher)


class TaskEffectiveResultCheck(AbstractMatchesTask):
    def _run(self, *args, **kwargs):
        self.asserts.check(self.parent.result.effective, self.matcher)


class TestEffectiveResultCheck(AbstractMatchesTask):
    def _run(self, *args, **kwargs):
        self.asserts.check(self.parent.result.effective, self.matcher)


class TestResultCheck(AbstractMatchesTask):
    def _run(self, *args, **kwargs):
        self.asserts.check(self.parent.result.current, self.matcher)


class AbstractExecResultMatchesTask(AbstractMatchesTask):
    @property
    def exec_result(self):
        if not self.context:
            return None
        return self.context.config['exec_result']



"""
Main cli module
"""
import json
import logging

import click

from ktdk.cli.manager import CliManager
from ktdk.cli.printer import print_tests
from ktdk.utils.flatters import flatten_tests

log = logging.getLogger(__name__)


@click.group(help='KTDK Cli tool')
@click.version_option('1.0')
def main_cli():
    pass


@main_cli.group('tests', help='Test management namespace')
def tests_cli():
    pass


@tests_cli.command('list', help='Lists all the tests')
@click.option('-f', '--full', default=False, is_flag=True, help='Show full test definition')
@click.option('-T', '--test-files', help='Test files location', required=False)
@click.option('-p', '--param', multiple=True,
              help="Override default parameters (format: name=value)")
def cli_tests_list(full, **kwargs):
    cli_manager = CliManager()
    ktdk = cli_manager.create_context(**kwargs)
    tests = flatten_tests(ktdk.suite)
    print("Number of tests: " + str(len(tests)))
    print_tests(tests, full=full)


@main_cli.command('execute', help='Executes the KTDK')
@click.option('--timeout', default=60, help='Timeout for each command to run')
@click.option('-D', '--devel', is_flag=True, default=False, help='Whether to run in devel mode')
@click.option('-T', '--test-files', help='Test files location', required=False)
@click.option('-S', '--submission', help='Submission files location', required=False)
@click.option('-W', '--workspace', help='Workspace location', required=False)
@click.option('-R', '--results', help='Results location', required=False)
@click.option('--clang-format-style', help='Clang format style default (pb161)', required=False)
@click.option('--clang-format-styles-dir', required=False,
              help='Clang format style templates dirs, (/tmp/style_check)')
@click.option('--dump-result', help='Dump result', required=False)
@click.option('-p', '--param', multiple=True,
              help="Override default parameters (format: name=value)")
def execute(param=None, **kwargs):
    params = {}
    if param:
        override = {item[0]: item[1] for item in param.split('=')}
        params.update(override)
    params.update({k: v for k, v in kwargs.items() if v is not None})
    log.info(f"[EXEC] Params: {params}")
    cli_manager = CliManager(**params)
    ktdk = cli_manager.create_context(**params)
    ktdk.invoke()
    stat = ktdk.stats
    print(json.dumps(stat, indent=2))


if __name__ == '__main__':
    main_cli()



import logging
import re

from ktdk.asserts import matchers
from ktdk.asserts.matchers import IsNotEmpty
from ktdk.core.tasks import Task
from ktdk.tasks.build_task import BuildTask

log = logging.getLogger(__name__)


class MakeResultParser(BuildTask):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.subdir = ''
        self.build_regex = re.compile(r"Built target (\w+)")

    def _run(self, *args, **kwargs):
        names = self.__parse_b_names()
        log.debug(f"[PRC] Binaries: {names}")
        self.asserts.check(names, IsNotEmpty("Binaries have not been found."))
        executables = {}
        for name in names:
            path = self.target / self.subdir / name
            executables[name] = path
        self.context.config.set_test('exec', executables)

    def __parse_b_names(self):
        names = []
        result = self.context.config['make_result']
        for line in result.stdout.content.splitlines():
            name = self.__parse_b_name(line)
            if name:
                names.append(name)
        return names

    def __parse_b_name(self, line):
        match = self.build_regex.search(line)
        return match.group(1) if match else None


class MakeTask(BuildTask):
    def __init__(self, **kwargs):
        super().__init__(command='make', **kwargs)
        self.args.extend(['-j', '4'])
        self.add_tags('make')

    def process_make(self, result):
        self.asserts.check(result, matchers.CommandOK())
        self.context.config.set_task('make_result', result)
        return result

    def run_make(self):
        result = self.execute()
        return self.process_make(result)

    def _run(self, *args, **kwargs):
        log.debug('[BLD] Make')
        self.move_to_dir(self.target)
        self.run_make()


class CMakeTask(BuildTask):
    def __init__(self, **kwargs):
        super().__init__(command='cmake', **kwargs)
        self.name = self.name or 'cmake'
        self.add_tags('cmake', 'compile')

    def run_cmake(self):
        self.args.append(str(self.source))
        result = self.execute()
        self.asserts.check(result, matchers.CommandOK())

    def _run(self, *args, **kwargs):
        log.debug('[BLD] CMake')
        self.move_to_dir(self.target)
        self.run_cmake()


class CMakeBuildTask(Task):
    def __init__(self, target='', source='', cwd=None, cmake_location=None, **kwargs):
        super().__init__(**kwargs)
        self.target = target
        self.source = source
        self.cmake_location = cmake_location
        self.name = 'cmake_build'
        self.cmake = CMakeTask(target_dir=target, source_dir=source, cwd=cwd)
        self.make = MakeTask(target_dir=target, source_dir=source, cwd=cwd)
        self.make.add_task(MakeResultParser(target_dir=target, source_dir=source))
        self.add_task(self.cmake)
        self.add_task(self.make)
        self._cwd = cwd



import logging
import os

from ktdk.tasks.command_task import CommandTask

log = logging.getLogger(__name__)


class BuildTask(CommandTask):
    def __init__(self, source_dir=None, target_dir=None, cwd=None, **kwargs):
        super().__init__(**kwargs)
        self._source_dir = source_dir or ''
        self._target_dir = target_dir or 'build'
        self._cwd = cwd
        self.add_tags('build')

    @property
    def source(self):
        return self.base / self._source_dir

    @property
    def base(self):
        return self.context.paths.workspace

    @property
    def target(self):
        return self.base / self._target_dir

    def source_dir(self, source):
        self._source_dir = source
        return self

    def target_dir(self, target):
        self._target_dir = target
        return self

    def __create_dir(self, directory=None, create=True):
        directory = directory or self.target
        if not directory.exists() and create:
            log.debug(f"[DIR] Create: {directory}")
            directory.mkdir(parents=True)

    def move_to_dir(self, directory, create=True):
        self.__create_dir(directory, create=create)
        os.chdir(directory)



#! /usr/bin/env python3

__author__ = "Peter Stanko"
__copyright__ = "Copyright 2016, The Project Name"
__credits__ = ["Peter Stanko"]
__license__ = "BSD"
__version__ = "0.0.1-SNAPSHOT"
__maintainer__ = "Peter Stanko"
__email__ = "wermington@gmail.com"
__status__ = "Production"




from typing import List

from ktdk import Test
from ktdk.utils.flatters import flatten_tasks


def print_tests(tests: List[Test], **kwargs):
    for i, test in enumerate(tests):
        print(f"{i + 1}: ", end='')
        if kwargs.get('full', False):
            print(test)
        else:
            print(f"{test.namespace} (P: {test.points} "
                  f") (T: {test.effective_tags})")
        print_tasks(test, test_counter=i, **kwargs)


def print_tasks(test: Test, test_counter=0, **kwargs):
    tasks = flatten_tasks(test)

    for i, task in enumerate(tasks):
        print(f"\t [TASK] {test_counter + 1}.{i+1}: ", end='')
        if kwargs.get('full', False):
            print(task)
        else:
            print(f"{task.namespace}")



from ktdk.core.tasks import Task
from tests.utils import TaskUtils, TestUtils


def assert_task(task, **params):
    params = {**TaskUtils.EMPTY, **params}
    assert params['name'] == task.name
    assert params['desc'] == task.desc


def test_empty_constructor():
    task = Task()
    assert_task(task)


def test_task_with_name_desc():
    task = Task(name='task01', desc='task01 desc')
    assert_task(task, name='task01', desc='task01 desc')


def test_with_all_constructor_params():
    params = TaskUtils.get_params()
    task = TaskUtils.create_task(**params)
    assert_task(task, **params)


def test_add_one_child_task():
    child_task = TaskUtils.create_task()
    task = TaskUtils.create_task()
    task.add_task(child_task)
    assert task.tasks
    assert child_task in task.tasks
    assert child_task.parent == task


def test_add_multiple_children_tasks():
    ch_tasks = [TaskUtils.create_task() for _ in range(5)]
    task = TaskUtils.create_task()
    task.add_task(*ch_tasks)
    for child in ch_tasks:
        assert child.parent == task
        assert child in task.tasks




from ktdk.runtime.context import Context


class ContextMixin(object):
    @property
    def context(self) -> Context:
        return getattr(self, '_context', None)

    def inject_context(self, context):
        setattr(self, '_context', context)


class CheckersMixin:
    def check_that(self, task: 'Task', points_multiplier=0, after_tasks=False):
        """Adds the check task
        Args:
            after_tasks:
            points_multiplier: If this task fails, how much from the test points should be discarded
            task(Task): Task instance
        """
        from .tasks import TaskType
        setattr(task, '_type', TaskType.CHECKED)
        setattr(task, '_points_multiplier', points_multiplier)
        if after_tasks:
            self.add_after(task)
        else:
            self.add_task(task)

    def require_that(self, task: 'Task', points_multiplier=0, after_tasks=False):
        """Requires a subtask to pass
        Args:
            points_multiplier: If this task fails, how much from the test points should be discarded
            task(Task):
        """
        from .tasks import TaskType
        self.check_that(task, points_multiplier=points_multiplier, after_tasks=after_tasks)
        setattr(task, '_type', TaskType.REQUIRED)



from ktdk import KTDK
from ktdk.runtime.runners import Runner

import pytest


@pytest.fixture()
def runner(mocker):
    runner = mocker.Mock(spec=Runner)
    return runner


def test_instance():
    assert KTDK.instance is None
    assert KTDK.get_instance() is not None
    assert KTDK.instance is not None


def test_config():
    ktdk = KTDK.get_instance()
    assert (ktdk.config['test_files'] == 'test_files')
    assert (ktdk.config['workspace'] == 'workspace')
    assert (ktdk.config['submission'] == 'submission')





from ktdk.core.tasks import Task


class DirLayoutActions(object):
    def __init__(self, task, source, subdir, destination=None):
        self.source = source
        self.destination = destination or 'workspace'
        self.task = task
        self.subdir = subdir or ''

    def init_task(self, task):
        task.source = self.source
        task.destination = self.destination
        task.subdir = self.subdir
        return task

    def add_task(self, *tasks):
        for task in tasks:
            self.init_task(task)
            self.task.add_task(task)

    def check_that(self, check):
        self.task.check_that(self.init_task(check))

    def require_that(self, check):
        self.task.require_that(self.init_task(check))


class FileTasks(Task):
    def _run(self, *args, **kwargs):
        pass

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = self.name or 'files'

    def workspace(self, subdir=None):
        return DirLayoutActions(task=self, subdir=subdir, source='workspace')

    def submission(self, subdir=None):
        return DirLayoutActions(task=self, subdir=subdir, source='submission')

    def test_files(self, subdir=None):
        return DirLayoutActions(task=self, subdir=subdir, source='test_files')



import re
import secrets
import string

from unidecode import unidecode

NOT_ALLOWED_FOR_NAME = string.punctuation.replace('_', '')


def slugify(text):
    """
    Simplifies ugly strings into something URL-friendly.
    >>> print slugify("[Some] _ Article's Title--")
    some-articles-title
    """

    # "[Some] _ Article's Title--"
    # "[some] _ article's title--"
    text = text.lower()

    # "[some] _ article's_title--"
    # "[some]___article's_title__"
    for c in [' ', '-', '.', '/']:
        text = text.replace(c, '_')

    # "[some]___article's_title__"
    # "some___articles_title__"
    text = re.sub('\W', '', text)
    # "some   articles title  "
    # "some articles title "
    text = re.sub('\s+', ' ', text)

    # "some articles title "
    # "some articles title"
    text = text.strip()

    # "some articles title"
    # "some-articles-title"
    text = text.replace(' ', '-')

    return text


def remove_punctuation(original: str) -> str:
    """Removes punctuation from string
    Args:
        original(str): Original string

    Returns(str): String with removed punctuation

    """
    table = original.maketrans('', '', NOT_ALLOWED_FOR_NAME)
    return original.translate(table)


def remove_diacritics(original: str) -> str:
    """Removes diacritics from the string
    Args:
        original(str): Original string

    Returns(str): String with removed diacritics

    """
    return unidecode(original)


def substitute_spaces(original: str, sub: str = '_') -> str:
    """Substitutes spaces
    Args:
        original(str): Original string
        sub(str): Substitute
    Returns(str): String with substituted spaces
    """
    return sub.join(original.split())


def convert_to_snake(original):
    """Converts an original string to snake case
    Args:
        original(str):  Original String

    Returns(str): Converted string

    """
    snaking = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', original)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', snaking).lower()


def normalize_name(name: str) -> str:
    """Gets a normalized name
    Args:
        name(str): Original name

    Returns(str): Normalized name
    """
    if name is None:
        return name
    name = convert_to_snake(name)
    name = slugify(name)
    name = remove_diacritics(name)
    name = name.lower()
    name = remove_punctuation(name)
    name = substitute_spaces(name)
    return name


def unique_suffix(length=8, context=None):
    if context and context.config['safe_suffixes'] is not True:
        return None
    return secrets.token_urlsafe(length)


def unique_name(name: str, context=None) -> str:
    """Gets an unique name
    Args:
        name(str): Original name
        context(AppContext): Optional context
    Returns(str): Unique name

    """
    suffix = unique_suffix(context=context)
    if suffix is None:
        return name
    return f"{name}_{suffix}"



"""
Runners for the test suite
"""
import abc
import logging

from ktdk.asserts import AssertionsChecks
from ktdk.core import results
from ktdk.core.errors import RequireFailedError, RequiredTaskFailed
from ktdk.utils.basic import BasicObject

log = logging.getLogger(__name__)


class Runner(BasicObject):
    def invoke(self, *args, **kwargs):
        self._run(*args, **kwargs)
        return self._process()

    @abc.abstractmethod
    def _run(self, *args, **kwargs):
        pass

    # pylint: disable=no-self-use
    def _process(self):
        return True
    # pylint: enable=no-self-use


class ExecutedTasksRegister(object):
    def __init__(self):
        self.collection = []

    def add_task(self, task):
        log.debug(f"[EXEC] Executed task: {task.namespace}")
        self.collection.append(task)

    def select(self, predicate=None):
        tasks = self.collection
        tasks = filter(predicate, tasks)
        return tasks

    def any(self, predicate=None):
        return any(self.select(predicate=predicate))


class TestRunner(Runner):
    def __init__(self, test, config, context):
        """Creates instance of the Test runner

        Args:
            test (Test): Test instance
            config (Dict): Runners configuration
            context (Context):
        """
        super().__init__()
        self.executed_tasks = ExecutedTasksRegister()
        self.test = test
        self.config = config
        self.context = context
        self.task_checks = []
        self.subtest_checks = []
        self.__init_test(context)

    # pylint: disable=no-self-use
    def _should_run(self):
        return self.context.tags.evaluate(*self.test.effective_tags)

    # pylint: enable=no-self-use

    def _run(self, *args, **kwargs):
        if not self._should_run():
            self.test.result = results.SKIP

        log.info(f"[RUN] Test: {self.test.namespace}")
        self._run_before_tasks()
        self._run_tasks()
        self._run_tests()
        self._run_after_tasks()

    def _process(self):
        return self.test

    def _run_tasks_collection(self, *args, collection=None, **kwargs):
        for task in collection:
            clone = self.context.clone()
            task_runner = task.runner.get_instance(
                context=clone,
                executed_tasks=self.executed_tasks
                )
            try:
                task_runner.invoke(*args, **kwargs)
            except RequiredTaskFailed as ex:
                log.info(f"[FAIL] Test [{self.test.namespace}] failed on task: {ex}")
                return False
        return True

    def _run_before_tasks(self):
        if self.test.parent:
            self._run_tasks_collection(collection=self.test.parent.before_each)
        return self._run_tasks_collection(collection=self.test.before)

    def _run_tasks(self):
        return self._run_tasks_collection(collection=self.test.tasks)

    def _run_after_tasks(self):
        if self.test.parent:
            self._run_tasks_collection(collection=self.test.parent.after_each)
        return self._run_tasks_collection(collection=self.test.after)

    def _run_tests(self):
        for child in self.test.tests:
            self._run_subtest(child)
        return all(test.result.effective.passed for test in self.test.tests)

    def _run_subtest(self, child):
        clone_context = self.context.clone(clone_test=True)
        child_runner = child.runner.get_instance(context=clone_context)
        child_runner.invoke()

    def __init_test(self, context):
        self.test.inject_context(context=context)


class TaskRunner(Runner):
    def __init__(self, task, config, context, executed_tasks=None):
        """Creates instance of the Task runner

       Args:
           task (Task): Task instance
           config (Dict): Runners configuration
           context (Context):
       """
        super().__init__()
        self.executed_tasks: ExecutedTasksRegister = executed_tasks or ExecutedTasksRegister()
        self.config = config
        self.context = context
        self.task = self.__init_task(task)

    # pylint: disable=no-self-use
    def _should_run(self):
        return True

    # pylint: enable=no-self-use

    def _run(self, *args, **kwargs):
        log.info(f"[RUN] Task: {self.task.namespace}")
        if not self._should_run():
            self.task.result = results.SKIP
        self.__execute_task_and_save_result()
        self._run_child_tasks()

    def __init_task(self, task):
        checks = AssertionsChecks()
        task.inject_checks(checks)
        task.inject_context(self.context)
        return task

    # pylint: disable=broad-except
    def __invoke_task_run(self):
        try:
            self.task.invoke()
        except RequireFailedError as ex:
            log.warning(f"[REQUIRE] Failed: {ex}")
        except Exception as ex:
            log.error(f"[ERROR] Task: {ex}")
            if self.context.devel:
                raise ex
            return results.ERROR
        if self.task.asserts.passed:
            return results.PASS
        return results.FAIL

    # pylint: enable=broad-except

    def __execute_task_and_save_result(self):
        result = self.__invoke_task_run()
        self.task.result.set(result=result)
        self.executed_tasks.add_task(self.task)

    def __halt_test_if_required(self, result):
        if not result.passed and self.task.required:
            raise RequiredTaskFailed(self.task)
        return result

    def _run_child_tasks(self, *args, **kwargs):
        for child in self.task.tasks:
            self._run_child_subtask(child=child, *args, **kwargs)

    def _run_child_subtask(self, child, *args, **kwargs):
        clone = self.context.clone()
        child_runner = child.runner.get_instance(
            context=clone,
            executed_tasks=self.executed_tasks
            )
        child_runner.invoke(*args, **kwargs)
        return child.result



from .basic import IsEmpty, IsFalse, IsNone, IsNotEmpty, IsTrue, Diff, \
    Contains, NotContains, NotEquals, IsNotNone, Equals, Greater, GreaterEquals, Less, LessEquals

from .suite import ChecksFailed, ChecksPassed, CommandOK, \
    CommandFailed, ResultPassed, ResultFailed, StderrIsEmpty, StdoutIsEmpty, StderrIsNotEmpty, \
    StdoutIsNotEmpty

from .general import GeneralExpectedMatcher, GeneralMatcher




import logging

from ktdk.core.tests import Test
from ktdk.utils.flatters import flatten_all_tasks, flatten_tests

log = logging.getLogger(__name__)


def _get_tests_with_result(tests, result):
    return [t for t in tests if getattr(t.result.current, result)]


def _count_for_tasks(test):
    tasks = flatten_all_tasks(test)
    return _count_for_collection(tasks)


def _count_for_collection(coll):
    count = dict(all=len(coll))
    for res_type in ['passed', 'failed', 'skipped', 'errored']:
        filtered = _get_tests_with_result(coll, res_type)
        count[res_type] = len(filtered)
    return count


def _get_tests_names(tests):
    result = {}

    def __names(res_type):
        sel = _get_tests_with_result(tests, res_type)
        if sel:
            result[res_type] = [selected.namespace for selected in sel]

    for rtype in ['failed', 'errored', 'skipped']:
        __names(res_type=rtype)
    return result


def _get_test_info(test):
    return dict(
        name=test.name,
        namespace=test.namespace,
        description=test.desc,
        tags=list(test.tags),
        effective_tags=list(test.effective_tags),
        points=test.points,
        result=dict(
            effective=test.result.effective.state,
            current=test.result.current.state,
            reduced_points=test.result.reduced_points,
            effective_points=test.result.effective_points
        )
    )


def _get_tests_info(all_tests):
    return [_get_test_info(test) for test in all_tests]


def stat_test(test) -> dict:
    """Statistics of the tests
    Args:
        test(Test): Root test

    Returns(dict): computed statistics

    """
    all_tests = flatten_tests(test)
    result = dict(final_points=test.result.effective_points,
                  result=test.result.effective.state)
    result['tests_count'] = _count_for_collection(all_tests)
    result['tasks_count'] = _count_for_tasks(test)
    result['nok_tests'] = _get_tests_names(all_tests)
    result['all_tests'] = _get_tests_info(all_tests)
    return result



import pytest

from ktdk import Context, Test
from ktdk.asserts.checks.general import TaskResultCheck
from ktdk.asserts.matchers import IsNone, IsNotNone, ResultPassed
from ktdk.core.tasks import Task


class FailingTask(Task):
    def _run(self, *args, **kwargs):
        self.asserts.check(None, IsNotNone())


class PassingTask(Task):
    def _run(self, *args, **kwargs):
        self.asserts.check(None, IsNone())


def get_failing_task():
    return FailingTask(name='failing_task')


def get_passing_task():
    return PassingTask(name='passing_task')


def get_failing_test(num: int = 0, points: float = 1):
    test = Test(name=f"test_{num}", desc=f"Failing test number {num}", points=points)
    fail = get_failing_task()
    test.add_task(fail)
    fail.check_that(TaskResultCheck(matcher=ResultPassed()))
    return test


def get_failing_reduce_test(num: int = 0, *reduces):
    test = Test(name=f"test_{num}", desc=f"Failing test number {num}", points=1)
    fail = get_failing_task()
    test.add_task(fail)
    for reduce in reduces:
        fail.check_that(TaskResultCheck(matcher=ResultPassed()), reduce)
    return test


def get_passing_test(num: int = 0, weight: float = 1):
    test = Test(name=f"test_{num}", desc=f"Passing test number {num}", points=weight)
    passing = get_passing_task()
    test.add_task(passing)
    passing.check_that(TaskResultCheck(matcher=ResultPassed()))
    return test


@pytest.fixture
def test():
    return Test(name='suite')


def test_should_have_full_points(test: Test):
    for i in range(10):
        test.add_test(get_passing_test(i + 1))
    runner = test.runner.get_instance(context=Context())
    runner.invoke()
    assert test.result.effective_points == 10


def test_should_have_half_points(test: Test):
    for i in range(10):
        test.add_test(get_passing_test(i + 1))
    for i in range(10):
        test.add_test(get_failing_test(i + 1))
    runner = test.runner.get_instance(context=Context())
    runner.invoke()
    assert test.result.effective_points == 10


def test_should_have_different_points_for_a_test(test: Test):
    test.add_test(get_passing_test(1))
    test.add_test(get_failing_test(1))
    test.add_test(get_failing_test(1, 4))
    test.add_test(get_passing_test(1, 2))

    runner = test.runner.get_instance(context=Context())
    runner.invoke()
    assert test.result.effective_points == 3


def test_should_have_multiple_layers(test: Test):
    naostro = get_passing_test(0, 5)
    nanecisto = get_passing_test(0, 3)

    naostro.add_test(get_passing_test(1))
    naostro.add_test(get_passing_test(3))
    naostro.add_test(get_failing_test(2))
    naostro.add_test(get_failing_test(4))

    nanecisto.add_test(get_passing_test(1))
    nanecisto.add_test(get_passing_test(3))
    nanecisto.add_test(get_failing_test(2))
    nanecisto.add_test(get_failing_test(4))

    test.add_test(naostro, nanecisto)

    runner = test.runner.get_instance(context=Context())
    runner.invoke()
    assert test.result.effective_points == 7 + 5
    assert naostro.result.effective_points == 7
    assert nanecisto.result.effective_points == 5


def test_should_have_multiple_passing_layers(test: Test):
    naostro = get_passing_test(0, 0.7)
    nanecisto = get_passing_test(0, 0.3)

    naostro.add_test(get_passing_test(1))
    naostro.add_test(get_passing_test(3))
    naostro.add_test(get_passing_test(2))
    naostro.add_test(get_passing_test(4))

    nanecisto.add_test(get_passing_test(1))
    nanecisto.add_test(get_passing_test(3))
    nanecisto.add_test(get_passing_test(2))
    nanecisto.add_test(get_passing_test(4))

    test.add_test(naostro, nanecisto)

    runner = test.runner.get_instance(context=Context())
    runner.invoke()
    assert test.result.effective_points == 9


def test_should_have_multiple_passing_layers_nanecisto_failing(test: Test):
    naostro = get_passing_test(0, 0.7)
    nanecisto = get_failing_test(0, 0.3)

    naostro.add_test(get_passing_test(1))

    nanecisto.add_test(get_passing_test(1))
    nanecisto.add_test(get_passing_test(3))
    nanecisto.add_test(get_passing_test(2))
    nanecisto.add_test(get_passing_test(4))

    test.add_test(naostro, nanecisto)

    runner = test.runner.get_instance(context=Context())
    runner.invoke()
    assert test.result.effective_points == 5.7


def test_should_reduce_points_if_test_fails(test: Test):
    test.add_test(get_failing_reduce_test(1, 0.3))
    runner = test.runner.get_instance(context=Context())
    runner.invoke()
    assert test.result.effective_points == 0.3


def test_should_reduce_multiple_points_if_test_fails(test: Test):
    test.add_test(get_failing_reduce_test(1, 0.5, 0.3))
    runner = test.runner.get_instance(context=Context())
    runner.invoke()
    assert test.result.effective_points == 0.3 * 0.5



"""
Tasks module
"""
import logging
from enum import Enum

from ktdk.asserts import AssertionMixin
from ktdk.core import results
from ktdk.core.abstract import GeneralObject, ResultHolder, RunnerConfig
from ktdk.core.mixins import CheckersMixin, ContextMixin
from ktdk.core.results import Result
from ktdk.runtime.runners import TaskRunner

log = logging.getLogger(__name__)


class TaskRunnerConfig(RunnerConfig):
    """Task runner configuration
    """

    @property
    def default_runner(self):
        """Gets a default runner for the task
        Returns:
        """
        return TaskRunner

    def __init__(self, task: 'Task', **params):
        """Creates a task runner
        Args:
            task(Task):
            **params:
        """
        super().__init__(**params)
        self.task = task

    # pylint: disable=arguments-differ
    def get_instance(self, **params):
        """Creates instance of the test runner
        Args:
            params: params
        Returns:
            TestRunner: TestRunner instance
        """
        return super(TaskRunnerConfig, self).get_instance(task=self.task, **params)
    # pylint: enable=arguments-differ


class TaskResultHolder(ResultHolder):
    """Keeps the results of the task, effective and current task
    """

    def __init__(self, task: 'Task'):
        super().__init__()
        self.task = task
        self._task_result = None
        self._subtasks_result: Result = None

    def reset_result_cache(self):
        super().reset_result_cache()
        self._subtasks_result = None
        self._task_result = None

    def set(self, result: Result):
        self._task_result = result

    @property
    def current(self) -> Result:
        return self._task_result or results.NONE

    @property
    def subtasks(self) -> Result:
        if self._subtasks_result is None:
            self._subtasks_result = self.__compute_subtasks_result()
        return self._subtasks_result

    def _compute_effective_result(self) -> Result:
        if not self.current.passed:
            return self.current
        return self.subtasks

    def __compute_subtasks_result(self) -> Result:
        if any(t.result.effective.errored for t in self.task.tasks):
            return results.ERROR
        if any(t.checked and t.result.effective.failed for t in self.task.tasks):
            return results.FAIL
        return results.PASS

    def to_dict(self):
        return {
            'type': 'task',
            'subtasks': self.subtasks.to_dict(),
            **super().to_dict(),
        }


class TaskType(Enum):
    """Defines Task types
    Types:
        - 'normal'
        - 'required'
        - 'checked'
    """
    NORMAL = 'normal'
    REQUIRED = 'required'
    CHECKED = 'checked'


class Task(GeneralObject, AssertionMixin, ContextMixin, CheckersMixin):
    """Task instance

    Attributes:
        desc(str): Description of the task
        name(str): Name of the task
        tasks(list): Child task that should be executed
        required(bool): Whether this task is required to be successful
    """

    def __init__(self, name: str = None, test=None, parent: 'Task' = None,
                 tasks: list = None, task_type: TaskType = None, **kwargs):
        """Creates instance of the task
        Args:
            name(str): Name of the task
            test(Test): Name of the test
            parent(Task): Parent task
            tasks(list): List of tasks
            **kwargs: Arguments provided to General Object
        """
        super().__init__(name=name, **kwargs)
        self._type = task_type or TaskType.NORMAL
        self.tasks: list = list(tasks or [])
        self._runner: TaskRunnerConfig = TaskRunnerConfig(self)
        self.__test = test
        self.__parent = parent
        self._result = TaskResultHolder(self)
        self._points_multiplier = 1

    @property
    def points_multiplier(self):
        """Gets a point multiplier
        Returns:

        """
        return self._points_multiplier

    @property
    def result(self) -> TaskResultHolder:
        """Gets a result for a Task
        Returns(TaskResultHolder):

        """
        return self._result

    @property
    def checked(self) -> bool:
        """Gets information whether task is checked or required or not
        Returns(bool): True if task is checked or required
        """
        return self.type in (TaskType.CHECKED, TaskType.REQUIRED)

    @property
    def required(self) -> bool:
        """Gets whether task is
        Returns:

        """
        return self.type == TaskType.REQUIRED

    @property
    def type(self) -> TaskType:
        """Gets type of the task
        """
        return self._type

    @property
    def effective_checked(self) -> bool:
        """Gets whether task is effective checked

        Task is effective checked, if the task itself is checked, or it's descendant tasks are
        checked.
        Returns(bool): true - if task is effective checked

        """
        if self.checked:
            return True
        if not self.tasks:
            return False
        return any(t.effective_checked for t in self.tasks)

    @property
    def test(self):
        """Gets an instance of the the owning test
        Returns(Test):
        """
        return self.__test

    @test.setter
    def test(self, value):
        """Sets an instance of the owning test
        Args:
            value(Test): Test instance
        """
        if self.__test is not None:
            log.warning(f'Overriding the test of the task: '
                        f'\"{self.name}\"; '
                        f'old: \"{self.test.namespace}\" '
                        f'new: \"{value.namespace}\"')
        setattr(self, '_log', None)
        self.__test = value

        for child in self.tasks:
            child.test = self.__test

    @property
    def parent(self) -> 'Task':
        """Gets an instance of the the parent task
        Returns(Task): Parent task instance
        """
        return self.__parent

    @parent.setter
    def parent(self, value: 'Task'):
        """Sets an instance of the the parent task
        Args:
            value(Task): Parent task instance
        """
        if self.parent is not None:
            log.warning(f'Overriding parent of the task: '
                        f'\"{self.name}\"; '
                        f'old: \"{self.parent.namespace}\" '
                        f'new: \"{value.namespace}\"')
        setattr(self, '_log', None)
        self.__parent = value

    @property
    def test_namespace(self) -> str:
        """Gets a test namespace
        Returns(str): Test namespace
        """
        return self.test.namespace if self.test else None

    @property
    def task_namespace(self) -> str:
        """Gets a task namespace
        Returns(str): Task namespace
        """
        name = []
        if self.parent is not None:
            name.append(self.parent.task_namespace)
        name.append(self.name)
        return ".".join(name)

    @property
    def namespace(self) -> str:
        """Gets a full namespace
        Returns(str): Full namespace
        """
        return ".".join([(self.test_namespace or ''), self.task_namespace])

    def add_task(self, *tasks: 'Task'):
        """Adds subtask
        Args:
            *tasks(Task):
        """
        self.__add_tasks(self.tasks, *tasks)

    def _run(self, *args, **kwargs):
        """Runs the task
        Args:
            *args: optional positional ars for run
            **kwargs: Keywords arguments for the run
        """
        pass

    def invoke(self, *args, **kwargs):
        """Invokes the task - runs and process the task
        Args:
            *args: optional positional ars for run
            **kwargs: Keywords arguments for the run
        """
        self._run(*args, **kwargs)
        return self._process()

    # pylint: disable=no-self-use
    def _process(self):
        """Process the task
        """
        return True

    # pylint: enable=no-self-use

    def to_dict(self) -> dict:
        """Converts the task to dictionary
        Returns(dict): task serialized do dictionary
        """
        res = {
            **(super().to_dict()),
            'result': self.result.to_dict(),
            'type': self.type.value,
        }
        if self.asserts:
            res['checks'] = self.asserts.to_dict()
        if self.tasks:
            res['tasks'] = [task.to_dict() for task in self.tasks]
        return res

    def __add_tasks(self, collection, *tasks):
        for task in tasks:
            task.test = self.test
            task.parent = self
            log.debug(f"[ADD] Task {self.namespace}: {task.name} [{task.namespace}]")
            collection.append(task)


class PipeTask(Task):
    def __init__(self, task=None, pipe=None, task_params=None, **kwargs):
        super().__init__(**kwargs)
        self.task = task or Task
        self.pipe = pipe
        self.task_params = task_params or {}

    def _run(self):
        params = {**self.task_params}
        params[self.pipe] = self.pipe_action()
        piped_task = self.task(**params)
        self.add_task(piped_task)

    def pipe_action(self):
        return None



from ktdk.core.tasks import Task
from ktdk.core.tests import Test as Instance, TestRunnerConfig
from ktdk.runtime.runners import TestRunner, Runner
from tests.utils import TestUtils, TaskUtils


def assert_test(test, **params):
    params = {**TestUtils.EMPTY, **params}
    assert params['name'] == test.name
    assert params['desc'] == test.desc
    assert set(params['tags']) == test.tags
    assert params['tests'] == test.tests
    assert params['before'] == test.before
    assert params['after'] == test.after


def test_empty_constructor():
    test = Instance()
    assert_test(test)


def test_with_name_and_desc():
    test = Instance(name='Test01', desc='Test desc')
    assert_test(test, name='test01', desc='Test desc')


def test_with_all_constructor_params():
    params = TestUtils.get_params()
    test = TestUtils.construct_test(**params)
    assert_test(test, **params)


def test_set_all_params_without_tasks_and_tests():
    params = TestUtils.get_params()
    test = TestUtils.create_test(**params)
    assert_test(test, **params)


def test_test_has_default_instance_of_runner():
    test = TestUtils.create_test()
    assert test.runner.runner == TestRunner


def test_updated_tags():
    params = TestUtils.get_params()
    test = TestUtils.create_test(**params)
    test.add_tags('new_tag')
    tags = {
        'tags': {'naostro', 'generic', 'stylecheck', 'new_tag'}
    }
    params = {**params, **tags}
    assert_test(test, **params)


def test_add_one_child_test():
    child_test = TestUtils.create_test()
    test = TestUtils.create_test(tests=[child_test])
    assert test.tests
    assert child_test in test.tests
    assert child_test.parent == test


def test_add_multiple_child_tests():
    ch_tests = [TestUtils.create_test() for _ in range(5)]
    test = TestUtils.create_test(tests=[*ch_tests])
    assert len(test.tests) == 5
    for child in ch_tests:
        assert child.parent == test
        assert child in test.tests


def test_add_tasks():
    tasks = [TaskUtils.create_task() for _ in range(5)]
    test = TestUtils.create_test(tasks=[*tasks])
    assert len(test.tasks) == 5
    for task in tasks:
        assert task.test == test
        assert task in test.tasks


def test_namespace_for_test_without_children():
    test = TestUtils.create_test(name="Test")
    assert test.namespace == "test"


def test_namespace_for_test_with_children():
    test = TestUtils.create_test(name='Test')
    child = TestUtils.create_test(name='Child')
    test.add_test(child)
    assert test.namespace == 'test'
    assert child.namespace == 'test::child'


def test_test_config_get_runner():
    test = TestUtils.create_test()
    runner = test.runner.get_instance()
    assert isinstance(runner, TestRunner)


def test_effective_tags():
    test = Instance(name='test', tags=['test'])
    naostro = Instance(name='naostro', tags=['naostro'])
    nanecisto = Instance(name='nanecisto', tags=['nanecisto'])
    test.add_test(naostro)
    test.add_test(nanecisto)
    assert 'test' in test.effective_tags
    assert 'naostro' in test.effective_tags
    assert 'nanecisto' in test.effective_tags
    assert 'nanecisto' not in naostro.effective_tags
    assert 'naostro' not in nanecisto.effective_tags
    assert 'test' in naostro.effective_tags
    assert 'test' in nanecisto.effective_tags



import json

import pytest

from ktdk import KTDK
from ktdk.asserts import matchers
from ktdk.asserts.checks.executable import *
from ktdk.asserts.checks.general import *
from ktdk.asserts.checks.style import StyleResultMatchesCheck
from ktdk.asserts.matchers import Equals, IsEmpty
from ktdk.core.tests import Test
from ktdk.tasks.command import Command
from ktdk.tasks.cpp.checkstyle import ClangFormatTask
from ktdk.tasks.cpp.cmake import CMakeBuildTask
from ktdk.tasks.cpp.valgrind import ValgrindCommand
from ktdk.tasks.fs.tasks import CopyFiles, ExistFiles, MakeDir
from ktdk.tasks.fs.tools import *
from ktdk.tasks.raw.executable import ExecutableTask, ValgrindExecutableTask


def create_file(where, name, content=None):
    default = (
        """#include <stdio.h>

int main() {
  printf("Hello %s!\\n");
  return 0;
}""" % name)
    content = content or default
    path = where.join(f'{name}.c')
    path.write(content)
    return path


def create_cmake(src, *names):
    content = """
    cmake_minimum_required (VERSION 3.0)
    project (kontr_tests)
    
    add_compile_options(-std=c99)

    
    """

    for name in names:
        content += f"add_executable({name} {name}.c)\n"
    cmake_file = src.join('CMakeLists.txt')
    cmake_file.write(content)


@pytest.fixture
def workspace_dir(tmpdir):
    workspace = tmpdir.mkdir('workspace')
    return workspace


@pytest.fixture
def test_files_dir(tmpdir):
    test_files = tmpdir.mkdir('test_files')
    return test_files


@pytest.fixture
def submission_dir(tmpdir):
    submission = tmpdir.mkdir('submission')
    return submission


@pytest.fixture()
def sources_dir(submission_dir):
    return submission_dir.mkdir('src')


@pytest.fixture()
def prepared_sources(sources_dir):
    create_file(sources_dir, 'main')
    create_file(sources_dir, 'ahoj')
    create_cmake(sources_dir, 'main', 'ahoj')
    return sources_dir


@pytest.fixture()
def hello_test():
    hello_test = Test(name="hello_test", desc="Super hello test", points=1)
    hello = ExecutableTask(executable='main', executor=Command)
    hello_test.add_task(hello)

    hello.check_that(ReturnCodeMatchesCheck(matcher=Equals(0)))
    hello.check_that(StdOutMatchesCheck(matcher=Equals("Hello main!\n")))
    hello.check_that(StdErrMatchesCheck(matcher=IsEmpty()))
    return hello_test


@pytest.fixture()
def ahoj_test():
    ahoj_test = Test(name="ahoj_test", desc="Super ahoj test", points=1)
    ahoj = ValgrindExecutableTask(executable='ahoj', executor=ValgrindCommand)
    ahoj_test.add_task(ahoj)

    ahoj.check_that(ReturnCodeMatchesCheck(matcher=Equals(0)))
    ahoj.check_that(ValgrindPassedCheck())
    ahoj.check_that(StdOutMatchesCheck(matcher=Equals("Hello ahoj!\n")))
    ahoj.check_that(StdErrMatchesCheck(matcher=IsEmpty()))
    return ahoj_test


@pytest.fixture()
def stylecheck_test():
    style_check_test = Test(name='style_check_test', desc='Style check test', points=0.5)
    clang_format = ClangFormatTask(file='src/main.c', style='Google')
    style_check_test.add_task(clang_format)
    clang_format.check_that(StyleResultMatchesCheck(matcher=IsEmpty()))
    return style_check_test


@pytest.fixture()
def naostro(hello_test, stylecheck_test):
    naostro = Test(name="naostro", desc="Test nanecisto", tags=['naostro'])
    naostro.add_test(hello_test)
    naostro.add_test(stylecheck_test)
    return naostro


@pytest.fixture()
def nanecisto(prepared_sources, ahoj_test):
    nanecisto = Test(name="nanecisto", desc="Test nanecisto", tags=['nanecisto'])
    nanecisto.add_test(ahoj_test)
    return nanecisto


@pytest.fixture()
def file_tasks():
    ft = FileTasks()
    ft.workspace('src').add_task(MakeDir())
    ft.submission('src').require_that(ExistFiles('main.c'))
    ft.submission('src').require_that(ExistFiles('ahoj.c'))

    ft.submission('src').add_task(CopyFiles('*.c'))
    ft.submission('src').add_task(CopyFiles('*.txt'))

    ft.workspace('src').check_that(ExistFiles("main.c"))
    ft.workspace('src').check_that(ExistFiles("ahoj.c"))
    ft.workspace('src').check_that(ExistFiles("CMakeLists.txt"))
    return ft


@pytest.fixture()
def ktdk(workspace_dir, test_files_dir, submission_dir, cmake, nanecisto, naostro, file_tasks):
    ktdk = KTDK(
        test_files=test_files_dir,
        submission=submission_dir,
        workspace=workspace_dir
    )
    ktdk.suite.add_task(file_tasks)
    ktdk.suite.add_task(cmake)
    ktdk.suite.add_test(nanecisto)
    ktdk.suite.add_test(naostro)
    return ktdk


@pytest.fixture
def cmake():
    cmake = CMakeBuildTask(source='src')
    cmake.require_that(TestResultCheck(matcher=matchers.ResultPassed()))
    return cmake


def test_ktdk_run(prepared_sources, ktdk):
    ktdk.invoke()
    assert ktdk.suite.result.effective.passed
    assert ktdk.suite.result.effective_points == 2.5



from ktdk import Test
from ktdk.asserts import checks, matchers
from ktdk.scenarios import FullScenario
from ktdk.tasks import fs
from ktdk.tasks.cpp import CMakeBuildTask, ValgrindCommand, catch


class CppMiniCatchCmakeFullScenario(FullScenario):
    def __init__(self, executables=None, points=3, executor=ValgrindCommand,
                 student_files=None, test_files=None, test_per_exec=False, **kwargs):
        super().__init__(**kwargs)
        self.executables = executables
        self.points = points
        self.executor = executor
        self.student_files = student_files or []
        self.test_files = test_files or []
        self._test_per_exec = test_per_exec

    @property
    def compute_points(self) -> catch.CatchCheckCasesAndComputePoint:
        return catch.CatchCheckCasesAndComputePoint(max_points=self.points)

    def catch_task(self, executable: str) -> catch.CatchRunTestsOneByOneTask:
        return catch.CatchRunTestsOneByOneTask(executable=executable, executor=self.executor)

    def file_tasks(self):
        for student in self.student_files:
            self.ft.submission().add_task(fs.CopyFiles(student))
        for test in self.test_files:
            self.ft.test_files().add_task(fs.CopyFiles(test))

    def compile_tasks(self):
        cmake = CMakeBuildTask()
        cmake.require_that(checks.TaskResultCheck(matcher=matchers.ResultPassed()))
        self.root_test.add_task(cmake)
        return cmake

    def run_tasks(self):
        for executable in self.executables:
            test = self.root_test
            if self._test_per_exec:
                test = Test(name=executable, desc=f'Subtest for exec: {executable}')
                self.root_test.add_test(test)
            test.add_task(self.catch_task(executable=executable))

    def evaluate_tasks(self):
        self.root_test.check_that(self.compute_points, after_tasks=True)



from .cmake import CMakeBuildTask, CMakeTask
from .valgrind import ValgrindCommand
from .compiler import RawCompilerTask, CCompilerTask, CPPCompilerTask
from .catch import CatchCheckCasesAndComputePoint, CatchRunTestsOneByOneTask, \
    CatchCheckAndComputePoint



import logging
from copy import deepcopy
from pathlib import Path
from typing import Union

from ktdk.runtime.tags import TagsEvaluator
from ktdk.utils.basic import BasicObject

log = logging.getLogger(__name__)


def deep_merge(source, destination):
    """Deep merge
    Examples:
    >>> a = { 'first' : { 'all_rows': { 'pass': 'dog', 'number': '1' } } }
    >>> b = { 'first' : { 'all_rows': { 'fail': 'cat', 'number': '5' } } }
    >>> deep_merge(b, a) == {'first':{'all_rows':{'pass': 'dog', 'fail': 'cat', 'number': '5'}}}
    True
    """
    for key, value in source.items():
        if isinstance(value, dict):
            # get node or create one
            node = destination.setdefault(key, {})
            deep_merge(value, node)
        else:
            destination[key] = value

    return destination


class ContextDirs(object):
    """Context dirs wrapper for paths
    """

    def __init__(self, context: 'Context'):
        """Creates instance of the paths dirs
        Args:
            context(Context): Context instance
        """
        self.context = context

    @property
    def workspace(self) -> Path:
        """Gets workspace path
        Returns(Path): Workspace path
        """
        return Path(self.context.config['workspace_dir'])

    @property
    def test_files(self) -> Path:
        """Test files path
        Returns(Path): Test file path
        """
        return Path(self.context.config['test_files_dir'])

    @property
    def submission(self) -> Path:
        """Submission files path
        Returns:

        """
        return Path(self.context.config['submission_dir'])

    def submission_path(self, sub_path: Union[str, Path], create=False) -> Path:
        return self.__get_subdir(self.submission, sub_path, create)

    def test_path(self, sub_path: Union[str, Path], create=False) -> Path:
        return self.__get_subdir(self.test_files, sub_path, create)

    def result_path(self, sub_path: Union[str, Path], create=False) -> Path:
        return self.__get_subdir(self.results, sub_path, create)

    def workspace_path(self, sub_path: Union[str, Path], create=False) -> Path:
        return self.__get_subdir(self.workspace, sub_path, create)

    def __get_subdir(self, where: Path, sub_path, create=False):
        full_path = where / Path(sub_path)
        if create and not full_path.exists():
            full_path.mkdir(parents=True)
        return full_path

    @property
    def results(self) -> Path:
        """Gets results path
        Returns(Path): Results path

        """
        results_dir = self.context.config['results_dir']
        if not results_dir:
            results_dir = self.workspace / 'results'
        return Path(results_dir)

    def get_dir(self, name) -> Path:
        return getattr(self, name)

    def resolve(self, name) -> Path:
        return getattr(self, name, name)

    def save_result(self, sub_path: Union[str, Path], content) -> Path:
        path = self.results / Path(sub_path)
        base_dir: Path = path.parent
        if not base_dir.exists():
            base_dir.mkdir(parents=True)
        log.debug(f"[SAVE] Save content to the results file: {sub_path}")
        path.write_text(content, encoding='utf-8')
        return path


class ContextConfig(object):
    def __init__(self, suite_config=None, test_config=None, task_config=None):
        self._suite = suite_config if suite_config is not None else {}
        self._test = test_config if test_config is not None else {}
        self._task = task_config if task_config is not None else {}

    @property
    def suite(self):
        return self._suite

    @property
    def test(self):
        return self._test

    @property
    def task(self):
        return self._task

    @property
    def all(self):
        config = {}
        config = deep_merge(self.suite, config)
        config = deep_merge(self.test, config)
        config = deep_merge(self.task, config)
        return config

    def __set_any(self, which, name, value):
        collection = getattr(self, which)
        if name in collection:
            log.debug(f"[CTX] Overriding: ({collection[name]}) by ({value})!")
        else:
            log.debug(f"[CTX] SET: {which}['{name}'] = {value}")
        collection[name] = value

    def __add_any(self, which, name, value):
        collection = getattr(self, which)
        old_value = collection.get(name, None)
        new_value = deepcopy(value)

        if isinstance(value, list):
            new_value = old_value + value if old_value else deepcopy(value)

        if isinstance(value, dict):
            new_value = deep_merge(old_value or {}, new_value)

        log.debug(f"[CTX] Adding ({which}): {old_value} + {value} -> {new_value}")
        collection[name] = new_value

    def add_suite(self, name, value):
        self.__add_any('suite', name, value)

    def add_test(self, name, value):
        self.__add_any('test', name, value)

    def add_task(self, name, value):
        self.__add_any('task', name, value)

    def set_suite(self, name, value):
        self.__set_any('suite', name, value)

    def set_test(self, name, value):
        self.__set_any('test', name, value)

    def set_task(self, name, value):
        self.__set_any('task', name, value)

    def clone(self, clone_test=False):
        test_config = deepcopy(self.test) if clone_test else self.test
        task_config = deepcopy(self.task)
        config = ContextConfig(suite_config=self.suite,
                               test_config=test_config,
                               task_config=task_config)
        return config

    def __getitem__(self, item):
        return self.all.get(item)

    def get(self, name):
        return self.all.get(name)

    def __str__(self):
        return str(self.all)

    def __repr__(self):
        return str(self.all)


class Context(BasicObject):
    def __init__(self, suite_config=None, test_config=None, task_config=None, config=None):
        super().__init__()
        self._config = config or ContextConfig(suite_config=suite_config,
                                               test_config=test_config,
                                               task_config=task_config)
        self._tags_evaluator = TagsEvaluator(self.config['tags'], self.config['registered_tags'])
        self._context_dirs = ContextDirs(self)

    @property
    def tags(self):
        return self._tags_evaluator

    @property
    def devel(self):
        return self.config.all.get('devel', False)

    @property
    def dirs(self) -> ContextDirs:
        return self._context_dirs

    @property
    def paths(self) -> ContextDirs:
        return self.dirs

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, value):
        self._config = value

    def clone(self, clone_test=False):
        return Context(config=self.config.clone(clone_test=clone_test))

    def to_dict(self):
        return self.config




class KTDKError(RuntimeError):
    pass


class KTDKAssertionError(KTDKError):
    pass


class RequireFailedError(KTDKAssertionError):
    pass


class RequiredTaskFailed(KTDKAssertionError):
    pass


class TaskRunFailed(KTDKError):
    pass



from ktdk.asserts.matchers.general import GeneralMatcher


class ChecksPassed(GeneralMatcher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='checks has passed')
        self.message = self.message or 'Checks did not passed'

    def predicate(self, orig):
        return orig.asserts.passed


class ChecksFailed(GeneralMatcher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='checks has failed')
        self.message = self.message or 'Checks did not failed'

    def predicate(self, orig):
        return orig.asserts.failed


class ResultPassed(GeneralMatcher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='result has passed')
        self.message = self.message or 'Result did not passed'

    def predicate(self, orig):
        return orig.passed


class ResultFailed(GeneralMatcher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='result has failed')
        self.message = self.message or 'Result did not failed'

    def predicate(self, orig):
        return orig.failed


class CommandOK(GeneralMatcher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='Command is OK')
        self.message = self.message or 'Command is not OK'

    def predicate(self, orig):
        return orig.ok


class CommandFailed(GeneralMatcher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='Command has Failed')
        self.message = self.message or 'Command has not Failed'

    def predicate(self, orig):
        return orig.ok


class StdoutIsEmpty(GeneralMatcher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='Stdout is not empty')
        self.message = self.message or 'Stdout is not empty'

    def message_footer(self, orig):
        return super().message_footer(orig=orig.stderr.content)

    def predicate(self, orig):
        return orig.stdout.empty


class StdoutIsNotEmpty(GeneralMatcher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='Stdout is empty')
        self.message = self.message or 'Stdout is empty'

    def message_footer(self, orig):
        return super().message_footer(orig=orig.stderr.content)

    def predicate(self, orig):
        return not orig.result.stdout.empty


class StderrIsEmpty(GeneralMatcher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='Stderr is not empty')
        self.message = self.message or 'Stderr is not empty'

    def message_footer(self, orig):
        return super().message_footer(orig=orig.stderr.content)

    def predicate(self, orig):
        return orig.result.stderr.empty


class StderrIsNotEmpty(GeneralMatcher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='Stderr is empty')
        self.message = self.message or 'Stderr is empty'

    def message_footer(self, orig):
        return super().message_footer(orig=orig.stderr.content)

    def predicate(self, orig):
        return not orig.result.stderr.empty



from ktdk.asserts import matchers
from ktdk.core.tasks import Task
from ktdk.tasks.command import Command


class CommandTask(Task):
    def __init__(self, command=None, args=None, executor=Command,
                 command_config=None, output_name=None, **kwargs):
        super().__init__(**kwargs)
        self.add_tags('command')
        self.name = self.name or command
        self.executor = executor or Command
        self._cmd = command
        self.command_config = command_config or {}
        self.args = args or []
        self._output_name = output_name

    def build_args(self, *args):
        arr = []
        arr.extend(args)
        arr.extend(self.args)
        return arr

    @property
    def command_path(self):
        return self._cmd

    def command(self):
        path = str(self.command_path)
        args = self.build_args()
        return self.executor(path, args=args, output_name=self._output_name, **self.command_config)

    def execute(self, **kwargs):
        return self.command().set_task(self).invoke(**kwargs)

    def _run(self, *args, **kwargs):
        result = self.execute()
        self.asserts.check(result, matchers.CommandOK())
        self.context.config.set_test('command_result', result)



import logging

from ktdk.core import results
from ktdk.core.abstract import GeneralObject, ResultHolder, RunnerConfig
from ktdk.core.mixins import CheckersMixin, ContextMixin
from ktdk.core.results import Result
from ktdk.runtime.runners import TestRunner
from ktdk.utils import flatters

log = logging.getLogger(__name__)


class TestRunnerConfig(RunnerConfig):
    @property
    def default_runner(self):
        return TestRunner

    def __init__(self, test, **params):
        super().__init__(**params)
        self.test = test

    # pylint: disable=arguments-differ
    def get_instance(self, **params):
        """Creates instance of the test runner
        Args:
            params: params
        Returns:
            TestRunner: TestRunner instance
        """
        return super(TestRunnerConfig, self).get_instance(test=self.test, **params)
    # pylint: enable=arguments-differ


class TestResultHolder(ResultHolder):
    def __init__(self, test: 'Test'):
        super().__init__()
        self.test = test
        self._tasks_result: Result = None
        self._subtests_result: Result = None
        self._effective_result: Result = None

    def reset_result_cache(self):
        self._tasks_result = None
        self._effective_result = None
        self._subtests_result = None

    @property
    def tasks(self) -> Result:
        if self._tasks_result is None:
            self._tasks_result = self.__compute_task_result()
        return self._tasks_result

    @property
    def current(self) -> Result:
        return self.tasks

    @property
    def subtests(self):
        if self._subtests_result is None:
            self._subtests_result = self.__compute_subtests_result()
        return self._subtests_result

    def _compute_effective_result(self) -> Result:
        if not self.tasks.passed:
            return self.tasks
        return self.subtests

    def __compute_subtests_result(self):
        if any(t.result.effective.errored for t in self.test.tests):
            return results.ERROR
        if any(t.result.effective.failed for t in self.test.tests):
            return results.FAIL
        return results.PASS

    def __compute_task_result(self):
        if any(task.result.effective.errored for task in self.test.tasks):
            return results.ERROR
        if any(task.result.effective.failed for task in self.test.tasks):
            return results.FAIL
        return results.PASS

    @property
    def checked_tasks(self):
        return [task for task in flatters.flatten_tasks(self.test) if task.checked]

    @property
    def failed_tasks(self):
        return [task for task in self.checked_tasks if task.result.effective.failed]

    @property
    def errored_tasks(self):
        return [task for task in self.checked_tasks if task.result.effective.errored]

    @property
    def passed_tasks(self):
        return [task for task in self.checked_tasks if task.result.effective.passed]

    @property
    def skipped_tasks(self):
        return [task for task in self.checked_tasks if task.result.effective.skipped]

    @property
    def nok_tasks(self):
        return self.errored_tasks + self.failed_tasks

    @property
    def reduced_points(self):
        return self.__reduce_points()

    @property
    def effective_points(self):
        return self.reduced_points + sum(test.result.effective_points for test in self.test.tests)

    def to_dict(self):
        return {
            'type': 'test',
            'subtests': self.subtests.to_dict(),
            'tasks': self.tasks.to_dict(),
            'effective_points': self.effective_points,
            'reduced_points': self.reduced_points,
            **super().to_dict()
        }

    def __reduce_points(self):
        points = self.test.points
        for task in self.nok_tasks:
            points *= task.points_multiplier
        return points


class Test(GeneralObject, CheckersMixin, ContextMixin):
    """Test instance

      Attributes:
          name(str): Name of the test
          desc(str): Description of the test
          tags(set): Test tags
          weight(float): Number of points for the test
      """

    def __init__(self, tests=None, tasks=None, weight: float = 1.0, points: float = 0,
                 before=None, after=None, before_each=None, after_each=None, **kwargs):
        """Creates instance of the Test (internal)

        Args:
            name(str): name of the test
            desc(str): description of the test
            tags(set): Collection of the tags
            tests(list): Collection of child tests
            tasks(list): Collection of checks for the test
            before(list): Tasks that will be executed before the test run
            after(list): Tasks that will be executed after the test run
            before_each(list): Tasks that will be executed before each child test
            after_each(list): Tasks that will be executed after each child test
        """
        super().__init__(**kwargs)
        self._parent = None

        self.add_task(*(tasks or []))
        self.add_before(*(before or []))
        self.add_after(*(after or []))

        self.add_before(*(before_each or []), scope='each')
        self.add_after(*(after_each or []), scope='each')
        self.add_test(*(tests or []))
        self._weight = weight
        self._points = points

        self._runner = TestRunnerConfig(self)
        self._result = TestResultHolder(self)

    @property
    def result(self) -> TestResultHolder:
        """Gets an instance of the result holder (wrapper over the test)
        Returns(TestResultHolder): Result holder for the test
        """
        return self._result

    def to_dict(self):
        """Converts test to dictionary representation
        Returns:

        """
        return \
            {
                **(super().to_dict()),
                'tags': list(self.tags),
                'effective_tags': list(self.effective_tags),
                'forward_tags': list(self.forward_tags),
                'backward_tags': list(self.backward_tags),
                'points': self.points,
                'before_tasks': [t.to_dict() for t in self.before],
                'tasks': [t.to_dict() for t in self.tasks],
                'tests': [t.to_dict() for t in self.tests],
                'after_tasks': [t.to_dict() for t in self.after],
                'result': self.result.to_dict()
            }

    @property
    def effective_tags(self) -> set:
        """Gets an effective tags set
        Returns(set): Effective tags set
        """
        return self.forward_tags.union(self.backward_tags)

    @property
    def forward_tags(self) -> set:
        """Gets an forward tags set
        Returns(set): Forward tags set
        """
        parent_tags = self.parent.tags if self.parent else []
        return self.tags.union(parent_tags)

    @property
    def backward_tags(self) -> set:
        """Gets an backward tags set
        Returns(set): Backward tags set
        """
        backward_set = set()
        for test in self.tests:
            backward_set = backward_set.union(test.effective_tags)
        backward_set = backward_set.union(self.tags)
        return backward_set

    @property
    def tests(self) -> list:
        """Gets tests collection
        Returns:
            List of tests
        """
        return getattr(self, '_tests', [])

    @property
    def tasks(self) -> list:
        """Gets tasks collection
        Returns:
            List of tasks
        """
        return getattr(self, '_tasks', [])

    @property
    def before(self) -> list:
        """Gets before task
        Returns(List[Task]):
            List of tasks
        """
        return getattr(self, '_before', [])

    @property
    def after(self) -> list:
        """Gets after tasks
        Returns(List[Task]):
            List of tasks
        """
        return getattr(self, '_after', [])

    @property
    def before_each(self) -> list:
        """Gets before each test tasks
        Returns(List[Task]):
            List of tasks
        """
        return getattr(self, '_before_each', [])

    @property
    def after_each(self) -> list:
        """Gets after each test tasks
        Returns(List[Task]):
            List of tasks
        """
        return getattr(self, '_after_each', [])

    @property
    def parent(self) -> 'Test':
        """Gets parent of the test
        Returns(Test):
            Test: Instance of the parent test
        """

        return self._parent

    @parent.setter
    def parent(self, value: 'Test'):
        """ Sets the parent class of the test

        Test will reset the _log and _report
        If test already has a parent, it will warn user of the class
        Args:
            value (Test): Parent of the test
        """
        if self.parent is not None:
            log.warning(f'Overriding parent of the test: '
                        f'\"{self.name}\"; '
                        f'old: \"{self.parent.namespace}\" '
                        f'new: \"{value.namespace}\"')
        setattr(self, '_log', None)
        self._parent = value

    @property
    def namespace(self) -> str:
        """Namespace of the test
        Returns(str): Full namespace of the test
        """
        name = []
        if self.parent is not None:
            name.append(self.parent.namespace)
        name.append(self.name)
        return "::".join(name)

    @property
    def points(self) -> float:
        return self._points

    @points.setter
    def points(self, value: float):
        self._points = value

    def add_task(self, *tasks):
        """ Adds the test checks
        Args:
            *tasks: Test checks that will be added to the test
        """
        self.__add_tasks('_tasks', *tasks)

    def add_test(self, *tests: 'Test'):
        """Adds child tests to the test
        Args:
            *tests: Collection of the child tests

        """
        if not hasattr(self, '_tests'):
            setattr(self, '_tests', [])

        for test in tests:
            test.parent = self
            log.debug(f"[REG] Adding test ({self.namespace}): {test.name} [{test.namespace}]")
            self.tests.append(test)

    def add_before(self, *tasks, scope='all'):
        """Adds before checks

        Before checks are checks that will be all executed before
        the actual test and its subtests
        If the scope is set to 'each',
        the checks will be executed before **each child test**

        Args:
            *tasks: Collection of the checks
            scope(str): Either 'all' or 'each' (default: 'all')
        """
        collection = '_before' if scope == 'all' else '_before_each'
        self.__add_tasks(collection, *tasks)

    def add_after(self, *tasks, scope='all'):
        """Adds after checks

        After checks are checks that will be all executed after
        the actual test and its subtests
        If the scope is set to 'each',
        the checks will be executed after **each child test**

        Args:
            *tasks: Collection of the checks
            scope: Either 'all' of 'each' (default: 'all')

        Returns:

        """
        collection = '_after' if scope == 'all' else '_after_each'
        self.__add_tasks(collection, *tasks)

    def __add_tasks(self, collection, *tasks):
        if not hasattr(self, collection):
            setattr(self, collection, [])
        col = getattr(self, collection)
        for task in tasks:
            task.test = self
            log.debug(f"[ADD] Task {self.namespace}: {task.name} [{task.namespace}]")
            col.append(task)



import os.path
from pathlib import Path
from typing import Optional, Union

from ktdk import Task
from ktdk.asserts.utils import get_context_diff
from ktdk.tasks.command_task import CommandTask


# clang-format --style=file "/path/to/file" | diff -B -rupN "/path/to/file" -
# https://clang.llvm.org/docs/ClangFormat.htmls

class ClangFormatFilesTask(Task):
    def __init__(self, files: list, style: str, command=None, **kwargs):
        self.command = command
        self.style = style
        self.files = files
        super().__init__(**kwargs)

    def _run(self, *args, **kwargs):
        for fpath in self.files:
            self.add_task(ClangFormatTask(file=fpath, style=self.style, command=self.command))


class ClangFormatTask(CommandTask):
    def __init__(self, file: Union[str, Path], style: str = None, command=None, **kwargs):
        command = command or 'clang-format'
        super().__init__(command=command, **kwargs)
        self.add_tags('checkstyle', 'format')
        self._file = file
        self._style = style

    @property
    def file(self) -> Path:
        if os.path.isabs(self._file):
            return Path(self._file)
        return self.context.paths.workspace_path(self._file)

    def get_diff_with_same_file(self, stdout: str):
        content = Path(self.file).read_text(encoding='utf-8')
        diff = get_context_diff(expected=content, provided=stdout)
        return diff

    def _run(self, *args, **kwargs):
        if self._style:
            self.args.extend(['--style', self._style])
        self.args.extend([str(self.file)])
        result = self.execute()
        diff = self.get_diff_with_same_file(stdout=result.stdout.content)
        diff_list = list(diff)
        diff_str = ''.join(diff_list)
        self.context.paths.save_result(f'diff/format_diff_{self.file.name}.diff', diff_str)
        self.context.config.set_test('command_result', result)
        self.context.config.set_test('tidy_result', diff_list)


# http://clang.llvm.org/extra/clang-tidy/
# clang-tidy -extra-arg-before=-x -extra-arg-before=c++
#   -extra-arg=-std=c++14
#   -header-filter='^(room.h)|(course.h)|(reservation.h)|(rsystem.h)|(handler.h)$'
#   main.cpp "handler.cpp" "rsystem.cpp" | sed -r 's#/([^/]+/)+_tmp_/([^/]+/)+##g'
class ClangTidyCheckStyle(CommandTask):
    def __init__(self, files: list, headers: list = None, command=None,
                 extra_args: list = None, extra_before=None, **kwargs):
        command = command or 'clang-tidy'
        super().__init__(command=command, **kwargs)
        self.add_tags('checkstyle', 'tidy')
        self._files = files
        self._headers = headers or []
        self._extra_args = extra_args or []
        self._extra_before = extra_before or []

    def file_path(self, fpath: str) -> str:
        if os.path.isabs(fpath):
            return str(fpath)
        return str(self.context.paths.workspace_path(fpath))

    @property
    def cwd(self) -> str:
        return str(self.context.paths.workspace)

    @property
    def files(self):
        return [self.file_path(fpath) for fpath in self._files]

    @property
    def headers(self) -> list:
        return self._headers

    @property
    def _header_filter(self) -> Optional[str]:
        if not self.headers:
            return None
        header_filter = "|".join([f"({header})" for header in self.headers])
        return f"\"^{header_filter}$\""

    @property
    def extra_args(self) -> list:
        return self._extra_args

    @property
    def extra_before(self) -> list:
        return self._extra_before

    def build_command_args(self) -> list:
        result = []
        for arg in self.extra_before:
            result.append(f'-extra-arg-before={arg}')
        for arg in self.extra_args:
            result.append(f'-extra-arg={arg}')
        header_filter = self._header_filter
        if header_filter:
            result.append(self._header_filter)
        result.extend(self.files)
        return result

    def _run(self, *args, **kwargs):
        self.args.extend(self.build_command_args())
        result = self.execute(cwd=self.cwd)
        self.context.config.set_test('command_result', result)
        self.context.config.set_test('tidy_result', result)

    def _process(self):
        super(ClangTidyCheckStyle, self)._process()


class CPPClangTidyCheckStyle(ClangTidyCheckStyle):
    def __init__(self, files: list, **kwargs):
        super().__init__(files, **kwargs)
        self.extra_args.extend(['-std=c++14'])
        self.extra_before.extend(['-x', 'c++'])


# http://clang.llvm.org/extra/clang-tidy/
# clang-tidy -extra-arg-before=-x -extra-arg-before=c++
#   -extra-arg=-std=c++14
#   -header-filter='^(room.h)|(course.h)|(reservation.h)|(rsystem.h)|(handler.h)$'
#   main.cpp "handler.cpp" "rsystem.cpp" | sed -r 's#/([^/]+/)+_tmp_/([^/]+/)+##g'
class CClangTidyCheckStyle(ClangTidyCheckStyle):
    def __init__(self, files: list, **kwargs):
        super().__init__(files, **kwargs)
        self.extra_args.extend(['-std=c99'])
        self.extra_before.extend(['-x', 'c'])



import abc
import logging
from json import JSONEncoder

from ktdk.core import results
from ktdk.core.results import Result
from ktdk.utils.basic import BasicObject
from ktdk.utils.naming import normalize_name

log = logging.getLogger(__name__)


class RunnerConfig(BasicObject):
    @property
    @abc.abstractmethod
    def default_runner(self):
        return None

    @property
    def default_config(self):
        return {}

    def __init__(self, runner=None, **params):
        super().__init__()
        self.runner = runner or self.default_runner
        self.config = {**self.default_config, **params}

    def get_instance(self, context=None, **params):
        """Creates instance of the runner
        Args:
            context(Context): Context fot the runner
        Returns:
            Runner: Runner instance
        """
        return self.runner(context=context, config=self.config, **params)

    def update_config(self, runner=None, **config):
        self.config = {**self.config, **config}
        self.update_runner(runner=runner)

    def update_runner(self, runner=None):
        self.runner = runner or self.default_runner


class GeneralObject(BasicObject):
    def __init__(self, name: str = None, desc: str = None, tags: set = None):
        super().__init__()
        self.name = name or self.__class__.__name__
        self._desc = desc
        self._tags = set(tags or {})
        self._runner = RunnerConfig()

    def to_dict(self) -> dict:
        """Converts the object to dictionary
        Returns(dict):
        """
        return {
            'name': self.name,
            'namespace': self.namespace,
            'description': self.desc,
            'tags': list(self.tags),
            }

    @property
    def runner(self):
        return self._runner

    @property
    def name(self):
        return getattr(self, '_name')

    @property
    def desc(self):
        return self._desc

    @desc.setter
    def desc(self, value):
        self._desc = value

    @name.setter
    def name(self, value):
        setattr(self, '_name', normalize_name(value))

    @property
    def namespace(self):
        return self.name

    def add_tags(self, *tags):
        self._tags = self._tags.union(tags)

    @property
    def tags(self):
        return self._tags


class ResultHolder(JSONEncoder):
    def __init__(self):
        super().__init__()
        self._effective_result: Result = None

    def default(self, o):  # pylint: disable=E0202
        return o.to_dict()

    @property
    def current(self) -> Result:
        return results.NONE

    @property
    def effective(self) -> Result:
        if self._effective_result is None:
            self._effective_result = self._compute_effective_result()
        return self._effective_result

    @abc.abstractmethod
    def _compute_effective_result(self) -> Result:
        return results.NONE

    def reset_result_cache(self):
        self._effective_result = None

    def to_dict(self):
        return {
            'type': 'task',
            'effective': self.effective.to_dict(),
            'current': self.current.to_dict(),
            }



from ktdk.asserts import AssertionsChecks, AssertionMixin
from ktdk.core.errors import RequireFailedError

import pytest


def test_assertion_support_assert_that():
    support = AssertionMixin()
    support.inject_checks(AssertionsChecks())
    assert support.asserts.check(1, lambda x: (True, x, 'must pass'))
    assert support.asserts.passed
    assert not support.asserts.check(2, lambda x: (False, x, 'must fail'))
    assert support.asserts.failed
    assert 1 in [i.obj for i in support.asserts.passed_results]
    assert 2 in [i.obj for i in support.asserts.failed_results]


def test_assertion_support_check_that():
    support = AssertionMixin()
    support.inject_checks(AssertionsChecks())

    assert support.asserts.check(1, lambda x: (True, x, 'must pass'))
    assert support.asserts.passed
    with pytest.raises(RequireFailedError):
        support.asserts.require(2, lambda x: (False, x, 'must fail'))
    assert support.asserts.failed
    assert 1 in [i.obj for i in support.asserts.passed_results]
    assert 2 in [i.obj for i in support.asserts.failed_results]


def test_assertion_support_require_that():
    support = AssertionMixin()
    support.inject_checks(AssertionsChecks())

    support.asserts.require(1, lambda x: (True, x, 'must pass'))
    assert support.asserts.passed
    assert not support.asserts.check(2, lambda x: (False, x, 'must fail'))
    assert support.asserts.failed
    assert 1 in [i.obj for i in support.asserts.passed_results]
    assert 2 in [i.obj for i in support.asserts.failed_results]




from ktdk.runtime.context import Context
from ktdk.tasks.fs.tasks import *

from ktdk.tasks.fs.tools import FileTasks


def prepare_dir(tmpdir, name):
    path = tmpdir.mkdir(name)
    for i in range(10):
        path.join(f"{name}_{i}.txt").write(f"{name}_{i}")
    return path


def create_context(tmpdir):
    config = dict(gc=0,
                  submission=prepare_dir(tmpdir, 'submission'),
                  workspace=prepare_dir(tmpdir, 'workspace'),
                  test_files=prepare_dir(tmpdir, 'test_files'))
    context = Context(config, {'lc': 0})
    return context


def test_files_copy_tasks(tmpdir):
    ft = FileTasks()
    ft.submission().add_task(CopyFiles('submission*'))
    ft.workspace().add_task(ExistFiles('submission*'))
    context = create_context(tmpdir=tmpdir)
    runner = ft.runner.get_instance(context=context)
    result = runner.invoke()
    assert result


def test_files_exist_tasks(tmpdir):
    ft = FileTasks()
    ft.workspace().add_task(ExistFiles('workspace*'))
    context = create_context(tmpdir=tmpdir)
    runner = ft.runner.get_instance(context=context)
    result = runner.invoke()
    assert result


def test_files_non_exist_tasks(tmpdir):
    ft = FileTasks()
    ft.workspace().check_that(ExistFiles('submission*'))
    context = create_context(tmpdir=tmpdir)
    runner = ft.runner.get_instance(context=context)
    result = runner.invoke()
    assert result


class CheckFoundFilesTask(Task):
    def _run(self, *args, **kwargs):
        self.asserts.check(self.context.config['found_files'], matcher=IsNotNone())


def test_files_find_task(tmpdir):
    ft = FileTasks()
    files = FindFiles('submission*')
    files.check_that(CheckFoundFilesTask())
    ft.workspace().add_task(files)
    context = create_context(tmpdir=tmpdir)
    runner = ft.runner.get_instance(context=context)
    result = runner.invoke()
    assert result



import pytest

from ktdk.tasks.cpp.compiler import CCompilerTask, RawCompilerTask
from tests.utils import get_test_context


@pytest.fixture
def prepared_dir(tmpdir):
    path = tmpdir.mkdir('workspace')
    src = path.mkdir('src')
    c_file = src.join('main.c')
    c_file.write("""
        #include <stdio.h>\n
        int main() { printf("ahoj svet!\\n"); return 0; }
    """)
    return path


@pytest.fixture
def context(prepared_dir):
    config = dict(workspace=prepared_dir, devel=True, c_compiler='clang')
    return get_test_context(suite_config=config, test_config={})


def test_compile_tasks(context):
    compiler = RawCompilerTask(compiler='gcc', executable='hello', files=['src/main.c'])
    runner = compiler.runner.get_instance(context=context)
    runner.invoke()
    assert compiler.result.effective.passed
    assert context.config['exec']


def test_compile_with_c_compiler_tasks(context):
    compiler = CCompilerTask(executable='hello', files=['src/main.c'])
    runner = compiler.runner.get_instance(context=context)
    runner.invoke()
    assert compiler.result.effective.passed
    assert context.config['exec']



import logging

from ktdk.core.errors import RequireFailedError
from ktdk.utils.basic import BasicObject

log = logging.getLogger(__name__)


class AssertionResult(BasicObject):
    """Defines assertion result object
    """
    def __init__(self, obj=None, message=None, **kwargs):
        super().__init__()
        self.message = message
        self.obj = obj
        self.other = kwargs

    def __str__(self):
        return f"{self.message}"


class AssertionsChecks(object):
    def __init__(self):
        self._passed = []
        self._failed = []

    def to_dict(self):
        return dict(
            passed_count=len(self._passed),
            failed=[str(r) for r in self._failed]
            )

    def __str__(self):
        return str(self.to_dict())

    def __add_any(self, what, obj=None, message=None, **kwargs):
        col = getattr(self, what)
        col.append(AssertionResult(obj=obj, message=message, **kwargs))

    def add_passed(self, obj=None, message=None, **kwargs):
        self.__add_any('_passed', obj=obj, message=message, **kwargs)

    def add_failed(self, obj=None, message=None, **kwargs):
        self.__add_any('_failed', obj, message, **kwargs)

    def check(self, obj, matcher=None):
        cond, _ = self.__check(obj, matcher=matcher)
        return cond

    def require(self, obj, matcher=None):
        cond, message = self.__check(obj, matcher=matcher)
        if not cond:
            raise RequireFailedError(message)

    def __check(self, obj, matcher):
        matcher = matcher or (lambda x: (x, x, x))
        condition, obj, message = matcher(obj)
        col = '_passed' if condition else '_failed'
        if not condition:
            log.info(f"[CHK] Failed: {message}")
        self.__add_any(col, obj, message)
        return condition, message

    @property
    def passed_results(self):
        return self._passed

    @property
    def failed_results(self):
        return self._failed

    @property
    def passed(self):
        return not self._failed

    @property
    def failed(self):
        return not self.passed


class AssertionMixin(object):
    def inject_checks(self, checks: AssertionsChecks):
        if not hasattr(self, '_asserts'):
            setattr(self, '_asserts', checks)
        else:
            log.debug(f"Resetting checks: {self.asserts} to {checks}")
            setattr(self, '_asserts', checks)

    @property
    def asserts(self) -> AssertionsChecks:
        return getattr(self, '_asserts', None)



from ktdk.tasks.command import Command


def test_command_construction():
    echo_cmd = "echo"
    command = Command(echo_cmd)
    assert command.command == echo_cmd
    assert command.args == []


def test_command_args():
    echo_cmd = ["echo"]
    command = Command(*echo_cmd)
    command.add_args("-Wall", "-Wextra")
    assert command.args == ["-Wall", "-Wextra"]


def test_execution_simple_echo():
    echo_cmd = ["ahoj svet!"]
    command = Command('echo', args=echo_cmd)
    result = command.run()
    assert result.return_code == 0
    assert result.stdout.content == "ahoj svet!\n"
    assert result.stderr.content == ""
    assert result.ok
    assert not result.timeout


def test_execution_return_non_zero():
    echo_cmd = "exit 1"
    command = Command(echo_cmd).shell(True)
    result = command.run()
    assert result.return_code == 1
    assert result.stdout.content == ""
    assert result.stderr.content == ""
    assert result.nok
    assert not result.timeout


def test_execution_timeout():
    echo_cmd = ["127.0.0.1"]
    command = Command('ping', args=echo_cmd)
    result = command.timeout(0.5).run()
    assert result.return_code != 0
    assert result.timeout
    assert result.nok



from ktdk.tasks.command import Command


class ValgrindCommand(Command):
    def __init__(self, command, args=None, **kwargs):
        """Creates Valgrind command instance
        Args:
            command(str): Command name
            args(list): Command arguments
        """
        super().__init__(command, args=args, **kwargs)
        self._vargs = []

    @property
    def valgrind_log_file(self):
        """Gets name of the valgrind log file
        Returns(str): Valgrind log file
        """
        if not hasattr(self, '__valgrind_log_file'):
            setattr(self, '__valgrind_log_file', self.output_file_name(suffix='valgrind.log'))
        return getattr(self, '__valgrind_log_file')

    @property
    def valgrind_command(self):
        """Valgrind command with arguments and sets log file
        Returns(list): valgrind command with valgrind arguments
        """
        args = ['valgrind']
        if self.context:
            file_path = self.context.paths.results.joinpath(self.valgrind_log_file)
            args.append(f'--log-file={file_path}')
        return ['valgrind',
                f'--log-file={file_path}',
                '--leak-check=full',
                *self._vargs]

    def add_vargs(self, *args):
        """Adds the valgrind arguments
        Args:
            *args: Valgrind arguments
        Returns(ValgrindCommand): Valgrind command instance
        """
        self._vargs.extend(args)
        return self

    @property
    def executable_command(self):
        """Executable command with arguments
        Returns(list): Executable command with arguments list
        """
        return [self.command, *self.args]

    @property
    def full_command(self):
        """Full command list
        Returns(list): Full command with valgrind and executable
        """
        return [*self.valgrind_command, *self.executable_command]

    def process(self, result):
        """Processes the result
        Args:
            result: Command execution result
        Returns: Processes result
        """
        process_result = super().process(result)
        process_result.outputs['valgrind_log'] = self.valgrind_log_file

        self.context.config.set_test('valgrind_log', self.valgrind_log_file)
        return process_result



"""
Basic object
"""
from json import JSONEncoder


class BasicObject(JSONEncoder):
    """ Base class from which all entities are going to inherit
    """

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return str(self.__class__.__name__) + ": " + str(self.default(o=self))

    def default(self, o):  # pylint: disable=method-hidden
        props = o.__dict__
        if hasattr(o, 'to_dict'):
            props = o.to_dict()
        return props



import pytest

from ktdk.core.tests import Test
from ktdk.tasks.xunit.junit import JUnitParseTask
from tests.utils import get_test_context

JUNIT_SUCCESS = """<?xml version="1.0" encoding="UTF-8"?>
<testsuites>
  <testsuite name="testbomb" errors="0" failures="0" tests="319" hostname="tbd" time="0.014154" timestamp="2018-03-02T09:00:01Z">
    <testcase classname="testbomb.global" name="Basic player actions/Move down, no obstacles there" time="0.000078"/>
    <testcase classname="testbomb.global" name="Basic player actions/Move left, stop on border" time="0.000053"/>
    <testcase classname="testbomb.global" name="Basic player actions/Drop a bomb" time="0.000028"/>
    <testcase classname="testbomb.Fixture" name="Game objects correctly implement visit() methods" time="0.000111"/>
    <testcase classname="testbomb.Fixture" name="Game object texture get/set" time="0.000045"/>
    <testcase classname="testbomb.Fixture" name="Scenario: It is possible to create obstacles/Given: An empty map/When: a brick is created/Then: a wall is on the map" time="0.000014"/>
    <testcase classname="testbomb.Fixture" name="Scenario: It is possible to create obstacles/Given: An empty map/When: A stone is created/Then: a wall is on the map" time="0.000011"/>
    <testcase classname="testbomb.Fixture" name="Scenario: Which objects blocks player movement?/Given: A player alongside a wall/When: The player goes towards the obstacle/Then: He does not move at all" time="0.000034"/>
    <testcase classname="testbomb.Fixture" name="Scenario: Which objects blocks player movement?/Given: A player alongside a bomb/When: The player goes towards the obstacle/Then: He does not move at all" time="0.000019"/>
    <testcase classname="testbomb.Fixture" name="Scenario: Which objects blocks player movement?/Given: A player alongside another player/When: The player goes towards the other player/Then: They end up standing on the same position" time="0.000021"/>
    <testcase classname="testbomb.Fixture" name="Scenario: Player can drop a bomb on an unaligned position" time="0.000122"/>
    <testcase classname="testbomb.Fixture" name="Scenario: Player can pick up a bonus/Given: Player standing next to a bonus item" time="0.000067"/>
    <testcase classname="testbomb.Fixture" name="Scenario: Player can land bombs/Given: A player with bomb limit = 1/When: The player lands a bomb/And when: The bomb explodes/Then: The player can land another bomb" time="0.000286"/>
    <testcase classname="testbomb.Fixture" name="Scenario: Player can land bombs/Given: Two players with bomb limit = 1/When: one player lands a bomb/Then: the other player can land a bomb, too" time="0.000009"/>
    <testcase classname="testbomb.Fixture" name="Scenario: Bomb explosion/Given: A newly planted bomb" time="0.000265"/>
    <testcase classname="testbomb.Fixture" name="Scenario: Bomb explosion/Given: A newly planted bomb/When: The time is almost out/Then: The bomb is still there" time="0.000003"/>
    <testcase classname="testbomb.Fixture" name="Scenario: Bomb explosion/Given: A newly planted bomb/When: The time is out/Then: The bomb is no longer there" time="0.000042"/>
    <testcase classname="testbomb.Fixture" name="Scenario: Bomb explosion/Given: A newly planted bomb/When: The time is out/Then: The bomb is no longer there/And: there is some fire around" time="0.000031"/>
    <testcase classname="testbomb.Fixture" name="Scenario: Bonus item does not stop fire/Given: A bomb with range=2, a bonus item nearby/When: the bomb explodes" time="0.000259"/>
    <testcase classname="testbomb.Fixture" name="Scenario: Bonus item does not stop fire/Given: A bomb with range=2, a bonus item nearby/When: the bomb explodes/Then: there is a fire on place of the bonus, and behind the bonus" time="0.000008"/>
    <testcase classname="testbomb.Fixture" name="Scenario: Brick does not fire/Given: A bomb with range=2, a brick wall nearby/When: the bomb explodes" time="0.000257"/>
    <testcase classname="testbomb.Fixture" name="Scenario: Brick does not fire/Given: A bomb with range=2, a brick wall nearby/When: the bomb explodes/Then: there is no fire neither on place of the wall nor behind the wall" time="0.00001"/>
    <testcase classname="testbomb.Fixture" name="Scenario: Bonus::BombMax/Given: A player with no bonus/When: the player eats Bonus::BombMax/And when: the player tries to land all 2 bombs to the same cell/Then: well, there is a bomb in the cell/And: but the player can still land another one bomb somewhere else" time="0.000013"/>
    <system-out/>
    <system-err/>
  </testsuite>
</testsuites>
"""

JUNIT_FAIL = """<?xml version="1.0" encoding="UTF-8"?>
<testsuites>
  <testsuite name="hw02_tests" errors="0" failures="17" tests="17" hostname="tbd" time="0.00183" timestamp="2018-03-02T13:50:27Z">
    <testcase classname="Read file, simple options" name="root" time="0.000017">
      <system-err>
Unable to open provided file inputs/basic.txt
Unable to open provided file empty
      </system-err>
    </testcase>
    <testcase classname="Read file, simple options" name="simple" time="0.00011">
      <failure message="3 == 0" type="REQUIRE">
at /tmp/hw02/solution/basic.cpp:16
      </failure>
    </testcase>
    <testcase classname="Read file, simple options" name="Not existing file" time="0.000046">
      <failure message="0 == 3" type="REQUIRE">
at /tmp/hw02/solution/basic.cpp:23
      </failure>
    </testcase>
    <testcase classname="Read file, simple options" name="Empty file" time="0.000048">
      <failure message="3 == 0" type="REQUIRE">
at /tmp/hw02/solution/basic.cpp:35
      </failure>
    </testcase>
    <testcase classname="Read file, options" name="root" time="0.00003">
      <system-err>
Unable to open provided file inputs/wrong1.txt
Unable to open provided file inputs/basic.txt
      </system-err>
    </testcase>
    <testcase classname="Read file, options" name="No permission" time="0.000037">
      <failure message="0 == 3" type="REQUIRE">
at /tmp/hw02/solution/basic.cpp:51
      </failure>
    </testcase>
    <testcase classname="Read file, options" name="Wrong: missing =" time="0.000064">
      <failure message="3 == 4" type="REQUIRE">
at /tmp/hw02/solution/basic.cpp:62
      </failure>
    </testcase>
    <testcase classname="Read file, options" name="Wrong: missing ]" time="0.000031">
      <failure message="3 == 4" type="REQUIRE">
at /tmp/hw02/solution/basic.cpp:66
      </failure>
    </testcase>
    <testcase classname="Read file, options" name="Wrong: Empty section name" time="0.000025">
      <failure message="3 == 4" type="REQUIRE">
at /tmp/hw02/solution/basic.cpp:70
      </failure>
    </testcase>
    <testcase classname="Read file, options" name="Wrong: Empty value name" time="0.000032">
      <failure message="3 == 4" type="REQUIRE">
at /tmp/hw02/solution/basic.cpp:74
      </failure>
    </testcase>
    <testcase classname="Read file, options" name="Wrong: Empty key name" time="0.000036">
      <failure message="3 == 4" type="REQUIRE">
at /tmp/hw02/solution/basic.cpp:78
      </failure>
    </testcase>
    <testcase classname="Read file, options" name="Wrong: key=value before first section" time="0.000037">
      <failure message="3 == 4" type="REQUIRE">
at /tmp/hw02/solution/basic.cpp:82
      </failure>
    </testcase>
    <testcase classname="Read file, options" name="Wrong: Empty value name" time="0.000043">
      <failure message="3 == 4" type="REQUIRE">
at /tmp/hw02/solution/basic.cpp:86
      </failure>
    </testcase>
    <testcase classname="Read file, options" name="Wrong: key=value before first section, not empty config" time="0.000037">
      <failure message="3 == 0" type="REQUIRE">
at /tmp/hw02/solution/basic.cpp:90
      </failure>
    </testcase>
    <testcase classname="Parsing basic file" name="root" time="0.000029">
      <system-err>
Unable to open provided file inputs/basic.txt
Unable to open provided file inputs/basic.txt
      </system-err>
    </testcase>
    <testcase classname="Parsing basic file" name="has Section" time="0.00002">
      <failure message="false" type="REQUIRE">
at /tmp/hw02/solution/basic.cpp:100
      </failure>
    </testcase>
    <testcase classname="Parsing basic file" name="has Key" time="0.000021">
      <failure message="false" type="REQUIRE">
at /tmp/hw02/solution/basic.cpp:107
      </failure>
    </testcase>
    <testcase classname="Parsing basic file" name="value" time="0.00003">
      <failure message="1 == 0" type="REQUIRE">
at /tmp/hw02/solution/basic.cpp:119
      </failure>
    </testcase>
    <testcase classname="Parsing basic file" name="isValue" time="0.000031">
      <failure message="false" type="REQUIRE">
at /tmp/hw02/solution/basic.cpp:137
      </failure>
    </testcase>
    <testcase classname="Parsing basic file" name="hasValue" time="0.000037">
      <failure message="false" type="REQUIRE">
at /tmp/hw02/solution/basic.cpp:149
      </failure>
    </testcase>
    <system-out/>
    <system-err>
Unable to open provided file inputs/basic.txt
Unable to open provided file empty
    </system-err>
  </testsuite>
</testsuites>
"""


def create_junit(out_dir, name, content):
    file_name = name + '.xml'
    out_dir = out_dir / 'results'
    junit = out_dir.join(file_name)
    junit.write(content)
    return str(junit)


@pytest.fixture()
def workspace(tmpdir):
    path = tmpdir.mkdir('workspace')
    path.mkdir('results')
    return path


@pytest.fixture()
def context(workspace):
    config = dict(workspace=workspace)
    return get_test_context(suite_config=config)


@pytest.fixture()
def succ_junit(workspace):
    return create_junit(workspace, 'succ', JUNIT_SUCCESS)


@pytest.fixture()
def fail_junit(workspace):
    return create_junit(workspace, 'fail', JUNIT_FAIL)


def test_should_parse_succ_junit(context, succ_junit):
    junit_file = succ_junit
    suites = JUnitParseTask(junit_file=junit_file)
    root_test = Test(name="root")
    root_test.add_task(suites)
    runner = root_test.runner.get_instance(context=context)
    runner.invoke()
    assert root_test.result.effective.passed


def test_should_parse_fail_junit(context, fail_junit):
    junit_file = fail_junit
    suites = JUnitParseTask(junit_file=junit_file)
    root_test = Test(name="root")
    root_test.add_task(suites)
    runner = root_test.runner.get_instance(context=context)
    runner.invoke()
    assert root_test.result.effective.failed



import pytest

from ktdk.tasks.cpp.checkstyle import ClangFormatTask
from tests.utils import get_test_context


@pytest.fixture
def prepared_dir(tmpdir):
    path = tmpdir.mkdir('workspace')
    c_file = path.join('main.cpp')
    c_file.write(
        """#include <iostream>

int main() {
  std::cout << "ahoj svet!" << std::endl;
  return 0;
}
""")
    return path


@pytest.fixture
def context(prepared_dir):
    config = dict(workspace=prepared_dir, devel=True)
    return get_test_context(suite_config=config, test_config={})


def test_clang_check_style_valid_task(context):
    checkstyle = ClangFormatTask(file='main.cpp', style='Google')
    runner = checkstyle.runner.get_instance(context=context)
    runner.invoke()
    assert checkstyle.result.effective.passed
    style_result = context.config['tidy_result']
    assert not style_result



from ktdk.asserts.matchers.general import GeneralExpectedMatcher, GeneralMatcher
from ktdk.asserts.utils import get_context_diff


class IsTrue(GeneralMatcher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='is')
        self.message = self.message or 'Is not true'

    def predicate(self, orig):
        return orig is True


class IsFalse(GeneralMatcher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='is')
        self.message = self.message or 'Is not false'

    def predicate(self, orig):
        return orig is False


class Equals(GeneralExpectedMatcher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='==')
        self.message = self.message or 'Does not equals'

    def predicate(self, orig):
        return orig == self.expected


class NotEquals(GeneralExpectedMatcher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='!=')
        self.message = self.message or 'Does equals'

    def predicate(self, orig):
        return orig != self.expected


class Greater(GeneralExpectedMatcher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='>')
        self.message = self.message or 'Is not Greater'

    def predicate(self, orig):
        return orig > self.expected


class Less(GeneralExpectedMatcher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='<')
        self.message = self.message or 'Is not Less'

    def predicate(self, orig):
        return orig < self.expected


class LessEquals(GeneralExpectedMatcher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='<=')
        self.message = self.message or 'Is not less equal'

    def predicate(self, orig):
        return orig <= self.expected


class GreaterEquals(GeneralExpectedMatcher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='>=')
        self.message = self.message or 'Is not greater equal'

    def predicate(self, orig):
        return orig >= self.expected


class IsNone(GeneralMatcher):
    def predicate(self, orig):
        return orig is None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='is None')
        self.message = self.message or 'Is not None'


class IsNotNone(GeneralMatcher):
    def predicate(self, orig):
        return orig is not None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='is not None')
        self.message = self.message or 'Is None'


class Contains(GeneralExpectedMatcher):
    def predicate(self, orig):
        return orig in self.expected

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='in')
        self.message = self.message or 'Does not contains'


class NotContains(GeneralExpectedMatcher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='not in', message='Does contains')
        self.message = self.message or 'Does contains'

    def predicate(self, orig):
        return orig not in self.expected


class IsNotEmpty(GeneralMatcher):
    def predicate(self, orig):
        return orig

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='is not Empty')
        self.message = self.message or 'Is empty'


class IsEmpty(GeneralMatcher):
    def predicate(self, orig):
        return not orig

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='is Empty')
        self.message = self.message or 'Is not empty'


class Diff(GeneralExpectedMatcher):
    def __init__(self, *args, keep_ends=True, diff_options=None, **kwargs):
        super().__init__(*args, **kwargs, symbol='diff')
        self.message = self.message or 'Diff is not empty - strings are not the same'
        self._diff_opts = diff_options
        self._keep_ends = keep_ends

    def invoke(self, orig):
        compare = get_context_diff(expected=self.expected, provided=orig, keep_ends=self._keep_ends)
        list_cmp = list(compare)
        cond = not list_cmp
        message = "\n" + ''.join(list_cmp)
        obj = self.get_object(orig)
        return cond, obj, message



import logging
from typing import Union

import junitparser
from junitparser import JUnitXml, TestCase, TestSuite

from ktdk.asserts.checks.general import TaskResultCheck
from ktdk.asserts.matchers import IsNone, ResultPassed
from ktdk.core.tasks import Task
from ktdk.core.tests import Test

log = logging.getLogger(__name__)


class _ProcessTestCaseTask(Task):
    def __init__(self, case: TestCase, **kwargs):
        super().__init__(**kwargs)
        self.case = case

    def _run(self, *args, **kwargs):
        self.asserts.check(self.case.result, matcher=IsNone())
        self.context.config.add_test('xunit_case', self.case)


def create_test(test_case: Union[TestCase, TestSuite], test: Test = None):
    test = Test() if test is None else test
    test.desc = test_case.name
    test.name = test_case.name
    return test


def build_test_case(test_case: TestCase):
    test = create_test(test_case=test_case)
    test.add_tags('junit_case')
    process_task = _ProcessTestCaseTask(test_case)
    process_task.check_that(TaskResultCheck(matcher=ResultPassed()))
    test.add_task(process_task)
    return test


class JUnitParseTask(Task):
    def __init__(self, junit_file=None, **kwargs):
        super().__init__(**kwargs)
        self._junit_file = junit_file
        self._root = None

    @property
    def xml_root(self):
        if self._root is None:
            self._root = JUnitXml.fromfile(self._junit_file)
        return self._root

    def parse_xunit(self):
        for suite in self.xml_root:
            self._handle_suite(suite)

    def _handle_suite(self, suite: TestSuite):
        log.debug(f"[HANDLE] SUITE: {suite}")
        self.add_task(_JUnitRunTheSuite(suite=suite))

    def _run(self, *args, **kwargs):
        log.debug(f"[JUNIT] Parse: {self.test.namespace}")
        return self.parse_xunit()


def _handle_case(test_case: TestCase, suite_test: Test):
    log.debug(f"[XUN] Test Case {test_case.name} [{test_case.result}]: {test_case}")
    case_test = build_test_case(test_case=test_case)
    suite_test.add_test(case_test)
    return suite_test


class _JUnitRunTheSuite(Task):
    def __init__(self, suite=None, **kwargs):
        super().__init__(**kwargs)
        self._suite: junitparser.TestSuite = suite

    @property
    def suite(self) -> junitparser.TestSuite:
        return self._suite

    def _run(self, *args, **kwargs):
        log.debug(f"[XUN] Test Suite {self.suite.name}: {self.suite}")
        self.context.config.set_test('xunit_suite', self.suite)
        suite_test = create_test(self.suite, test=self.test)
        suite_test.add_tags('junit_suite')
        for case in self.suite:
            _handle_case(case, suite_test=suite_test)
        return suite_test



from .general import TaskResultCheck, TaskEffectiveResultCheck, \
    TestResultCheck, TestEffectiveResultCheck

from .executable import ReturnCodeMatchesCheck, StdErrMatchesCheck, \
    StdOutMatchesCheck, ValgrindPassedCheck


from .junit import XUnitSuiteErrorsCheck, XUnitSuiteFailsCheck, XUnitSuiteNokCheck, XUnitCaseCheck


from .style import ClangTidyResultMatchesCheck, StyleResultMatchesCheck



from ktdk.asserts import matchers
from ktdk.asserts.matchers import Equals, IsEmpty
from ktdk.asserts.checks.executable import *
from ktdk.asserts.checks.general import *
from ktdk.core.tests import Test
from ktdk import KTDK

from ktdk.tasks.cpp.cmake import CMakeBuildTask
from ktdk.tasks.cpp.valgrind import ValgrindCommand
from ktdk.tasks.fs.tasks import MakeDir, ExistFiles, CopyFiles
from ktdk.tasks.fs.tools import *
from ktdk.tasks.raw.executable import ExecutableTask


ktdk = KTDK.get_instance()

naostro = Test(name="naostro", desc="Test nanecisto")
ktdk.suite.add_test(naostro)

ft = FileTasks()
ft.workspace('src').add_task(MakeDir())
ft.workspace('results').add_task(MakeDir())
ft.submission('src').require_that(ExistFiles('main.c'))
ft.test_files().require_that(ExistFiles('CMakeLists.txt'))

ft.submission('src').add_task(CopyFiles('*.c'))
ft.test_files().add_task(CopyFiles('CMakeLists.txt'))

ft.workspace('src').check_that(ExistFiles("main.c"))

ktdk.suite.add_task(ft)


cmake = CMakeBuildTask()
cmake.check_that(TestResultCheck(matcher=matchers.ResultPassed()))
ktdk.suite.add_task(cmake)

hello_test = Test(name="hello_test", desc="Super hello test")
hello = ExecutableTask(executable='hello', executor=ValgrindCommand)
hello_test.add_task(hello)


hello.check_that(ReturnCodeMatchesCheck(matcher=Equals(0)))
hello.check_that(StdOutMatchesCheck(matcher=Equals("Hello world!")))
hello.check_that(StdErrMatchesCheck(matcher=IsEmpty()))

naostro.add_test(hello_test)




from typing import List


def flatten_collection(obj, collection: str, include_self=True):
    """Flattens any collection (tree) to list
    Args:
        include_self: Whether to include the root
        obj: Root element
        collection(str): Collection name (example: tests, tasks)
    Returns(List): List of all elements in the tree
    """
    if obj is None:
        return []
    result = [obj] if include_self else []
    for t in getattr(obj, collection):
        result.extend(flatten_collection(t, collection))
    return result


def flatten_tests(test, include_self: bool = True) -> List:
    """Flattens all the tests
    Args:
        include_self(bool): Whether to include root test
        test(Test): Root test
    Returns(List[Test]): List of all tests
    """
    if test is None:
        return []
    result = [test] if include_self else []
    for t in test.tests:
        result.extend(flatten_tests(t))
    return flatten_collection(test, 'tests', include_self=include_self)


def flatten_tasks(test) -> List:
    """Flatten all the tasks in the tests
    Args:
        test(Test): Root test
    Returns(List[Task]): Tasks
    """
    result = []
    for task in test.tasks:
        result.extend(flatten_collection(task, 'tasks'))
    return result


def flatten_all_tasks(test) -> List:
    """Flatten all the tasks in all the tests
    Args:
        test(Test): Root test
    Returns(List[Task]): Tasks
    """
    result = []
    all_tests = flatten_tests(test=test)
    for selected_test in all_tests:
        result.extend(flatten_tasks(selected_test))
    return result



import logging

from ktdk.asserts import matchers
from ktdk.tasks.command_task import CommandTask
from ktdk.tasks.cpp.valgrind import ValgrindCommand

log = logging.getLogger(__name__)


class ExecutableTask(CommandTask):
    def __init__(self, executable, *args, **kwargs):
        super().__init__(command=None, args=args, **kwargs)
        self.executable_name = executable

    @property
    def command_path(self):
        executables = self.context.config['exec']
        if not executables:
            log.error(f"[EXEC] Executables has not been found in the context")
        self.asserts.require(executables, matcher=matchers.IsNotNone())
        executable = executables.get(self.executable_name)
        if not executable:
            log.error(f"[EXEC] Executable not exists {self.executable_name}")
        self.asserts.require(executable, matcher=matchers.IsNotNone())
        return executable

    def _run(self, *args, **kwargs):
        result = self.execute()
        self.context.config.set_test('exec_result', result)


class ValgrindExecutableTask(ExecutableTask):
    def __init__(self, executable, *args, executor=ValgrindCommand, **kwargs):
        super().__init__(executable, *args, executor=executor, **kwargs)



import re

from ktdk.asserts.checks.general import AbstractExecResultMatchesTask


class ReturnCodeMatchesCheck(AbstractExecResultMatchesTask):
    def _run(self, *args, **kwargs):
        self.asserts.check(self.exec_result.return_code, self.matcher)


class StdOutMatchesCheck(AbstractExecResultMatchesTask):
    def _run(self, *args, **kwargs):
        self.asserts.check(self.exec_result.stdout.content, self.matcher)


class StdErrMatchesCheck(AbstractExecResultMatchesTask):
    def _run(self, *args, **kwargs):
        self.asserts.check(self.exec_result.stderr.content, self.matcher)


class ValgrindPassedCheck(AbstractExecResultMatchesTask):
    REGEX = re.compile(r'ERROR SUMMARY: (\d+) errors')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def points_multiplier(self):
        if self._points_multiplier:
            return self._points_multiplier
        else:
            return self.context.config['valgrind_penalization'] or 0

    @property
    def valgrind_log(self) -> str:
        return self.context.config['valgrind_log']

    def __read_log_content(self):
        valgrind_log = self.valgrind_log
        if valgrind_log is None:
            return None
        valgrind_log = self.context.paths.results / valgrind_log
        with open(valgrind_log, 'r') as content_file:
            return content_file.read()

    def _run(self, *args, **kwargs):
        self.asserts.check(self._asserted_zero_errors(), self.matcher)

    def _asserted_zero_errors(self):
        content = self.__read_log_content()
        if content is None:
            return False
        for line in content.splitlines():
            match = ValgrindPassedCheck.REGEX.search(line)
            if match:
                errors = match.group(1)
                return int(errors) == 0
        return True



import logging
import os
import shutil
from pathlib import Path

from ktdk.asserts.matchers import IsNotEmpty, IsNotNone, IsTrue
from ktdk.core.tasks import PipeTask, Task

log = logging.getLogger(__name__)


def copytree(src, dst, symlinks=False, ignore=None):
    for item in os.listdir(src):
        source = os.path.join(src, item)
        destination = os.path.join(dst, item)
        if os.path.isdir(source):
            shutil.copytree(source, destination, symlinks, ignore)
        else:
            shutil.copy2(source, destination)


class FileActionWrapper(object):
    def __init__(self, action, path):
        self.action = action
        self.path = path

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return self.__dict__


class AbstractFilesTask(Task):
    def __init__(self, pattern='', subdir='', from_dir=None, to_dir=None, output_subdir=None,
                 source='submission', destination='workspace',
                 **kwargs):
        super().__init__(**kwargs)
        self.source = source
        self.destination = destination
        self.subdir = subdir or ''
        self.output_subdir = output_subdir or ''
        self._from_dir = Path(from_dir) if from_dir is not None else None
        self._to_dir = Path(to_dir) if to_dir is not None else None
        self.pattern = [pattern] if isinstance(pattern, str) else pattern

    @property
    def from_path(self) -> Path:
        if self._from_dir is not None:
            return self._from_dir
        return self.context.dirs.get_dir(self.source) / self.subdir

    @property
    def to_path(self):
        if self._to_dir is not None:
            return self._from_dir
        return self.context.dirs.get_dir(self.destination) / self.subdir / self.output_subdir

    def _glob_files(self, path=None):
        path = path or self.from_path
        found = []
        for pat in self.pattern:
            found += path.glob(pat)
        return found

    def _glob_dirs(self, path=None):
        path = path or self.from_path
        found = []
        if self.pattern:
            for pat in self.pattern:
                found += path.glob(pat)
            return found
        return [path]

    def _process_files(self, subdir='', transform_path=None, action=None, **kwargs):
        found = self._glob_dirs()
        path = self.to_path
        final_dir = path.joinpath(subdir)
        for i in found:
            if transform_path:
                i = transform_path(i)
            final_path = str(final_dir.resolve())
            if final_dir.is_dir():
                final_path += "/"
            if final_dir.exists():
                log.debug(f"[FSM] Manip ({action.__name__}): {i} -> {final_path}")
                self.context.config.add_suite('work_files', [str(i)])
                action(i, final_path, **kwargs)
            else:
                log.error(f"[FILE] Cannot find file path: {final_dir}")


class CopyFiles(AbstractFilesTask):
    def _run(self, *args, **kwargs):
        self._process_files(action=shutil.copy2, **kwargs)


class CopyDir(AbstractFilesTask):
    def _run(self, *args, **kwargs):
        self._process_files(action=copytree, **kwargs)


class MoveFiles(AbstractFilesTask):
    def _run(self, *args, **kwargs):
        self._process_files(action=shutil.move, transform_path=str, **kwargs)


class DeleteFiles(AbstractFilesTask):
    def _run(self, *args, **kwargs):
        files = self._glob_files(path=self.to_path)
        for file_to_delete in files:
            log.debug(f"[DEL] Deleting file: {file_to_delete}")
            file_to_delete.unlink()


class ExistFiles(AbstractFilesTask):
    def _run(self, *args, **kwargs):
        files = self._glob_files(path=self.from_path)
        self.asserts.check(
            files,
            IsNotEmpty(message=f'Files not found! Pattern {self.pattern} in {self.from_path}')
        )
        exist = all(f.exists() for f in files)
        message = f"Not found files for pattern: {self.pattern} in {self.from_path}"
        self.asserts.check(exist, IsTrue(message=message))


class FindFiles(AbstractFilesTask):
    def _run(self, *args, **kwargs):
        files = self._glob_files(path=self.from_path)
        log.debug(f"[FND] Found files ({self.from_path}/{self.pattern}): {files}")
        self.context.config.add_task('found_files', files)
        self.asserts.check(files, IsNotNone(message=f"Cannot find any files: {self.pattern}"))


class MakeDir(AbstractFilesTask):
    def _run(self, *args, **kwargs):
        log.debug(f"[DIR] Create: {self.from_path}")
        self.from_path.mkdir(parents=True, exist_ok=True)


class ProcessFiles(AbstractFilesTask):
    def __init__(self, pattern, action=None, **kwargs):
        super().__init__(pattern, **kwargs)
        self.action = action

    def _run(self, *args, **kwargs):
        self._process_files(action=self.action)


class FileGlobPipeTask(PipeTask):
    def __init__(self, pattern, **kwargs):
        super().__init__(**kwargs)
        self.pattern = pattern

    def pipe_action(self):
        return self.context.paths.workspace.glob(self.pattern)



import difflib


def get_context_diff(expected, provided, keep_ends=True):
    def _split_string(orig):
        if isinstance(orig, str):
            orig = orig.splitlines(keepends=keep_ends)
        return orig

    expected = _split_string(expected)
    provided = _split_string(provided)
    return difflib.context_diff(expected, provided, fromfile='expected', tofile='provided')



import pytest

from ktdk.utils.basic import BasicObject


@pytest.fixture()
def basic():
    return BasicObject()


def test_basic_object_to_string(basic):
    assert 'BasicObject: ' in basic.__str__()


def test_basic_object_repr(basic):
    assert basic.__repr__() == basic.__str__()



import pytest

from ktdk.runtime.context import Context


@pytest.fixture()
def context():
    return Context(
        suite_config={'global': 0},
        test_config={'test': 10},
        task_config={'task': 100}
    )


def test_context_init():
    context = Context({})
    assert not context.config.all


def test_context_config(context):
    assert context.config['global'] == 0
    assert context.config['test'] == 10
    assert context.config['task'] == 100


def test_context_clone_no_test(context):
    context.config.set_suite('global', 1)
    context.config.set_test('test', 11)
    context.config.set_task('task', 101)

    assert context.config['global'] == 1
    assert context.config['test'] == 11
    assert context.config['task'] == 101
    clone = context.clone()
    assert id(context.config.test) == id(clone.config.test)
    assert clone.config['global'] == 1
    assert clone.config['test'] == 11
    assert clone.config['task'] == 101
    clone.config.set_task('task', 102)
    assert clone.config['task'] == 102
    clone.config.set_test('test', 12)
    assert context.config['task'] == 101
    assert context.config['test'] == 12


def test_context_clone_test(context):
    context.config.set_suite('global', 1)
    context.config.set_test('test', 11)
    context.config.set_task('task', 101)

    assert context.config['global'] == 1
    assert context.config['test'] == 11
    assert context.config['task'] == 101
    clone = context.clone(clone_test=True)
    assert clone.config['global'] == 1
    assert clone.config['test'] == 11
    assert clone.config['task'] == 101
    clone.config.set_task('task', 102)
    assert clone.config['task'] == 102
    clone.config.set_test('test', 12)
    assert context.config['task'] == 101
    assert context.config['test'] == 11


def test_context_clone_test2(context):
    task_clone = context.clone()
    task_clone.config.set_test('test', 1000)
    assert context.config['test'] == 1000
    clone = context.clone(clone_test=True)
    assert clone.config['test'] == 1000
    clone.config.set_test('test', 12)
    assert clone.config['test'] == 12
    clone2 = clone.clone(clone_test=True)
    assert context.config['test'] == 1000
    assert clone2.config['test'] == 12


def test_context_add_test_dict(context):
    context.config.set_task('comp', {'hello': 'world'})
    context.config.add_task('comp', {'hi': 'sky'})
    assert 'hello' in context.config['comp']
    assert 'hi' in context.config['comp']
    assert 'world' == context.config['comp']['hello']
    assert 'sky' == context.config['comp']['hi']


def test_context_add_test_list(context):
    context.config.set_task('comp', ['hello'])
    context.config.add_task('comp', ['hi'])
    assert 'hello' in context.config['comp']
    assert 'hi' in context.config['comp']



import pytest

from ktdk.asserts import AssertionsChecks
from ktdk.asserts.matchers import *
from ktdk.core import results
from ktdk.core.tasks import Task


@pytest.fixture
def task():
    task = Task('test_task')
    task.inject_checks(AssertionsChecks())
    return task


def test_general_matcher_invoke():
    matcher = GeneralMatcher('basic matcher')
    cond, obj, message = matcher.invoke("some string")
    assert cond
    assert obj == "some string"


def test_general_expected_matcher_invoke():
    matcher = GeneralExpectedMatcher('expected string')
    cond, obj, message = matcher.invoke('expected string')
    assert cond
    assert matcher.expected == "expected string"
    assert obj == "expected string"


def test_matchers_diff():
    expected = "Ahoj svet ako sa mas?\nJa sa mam dobre.\nSame"
    orig = "Ahoj svet ako sa mas?\nJa sa mam dobre.\nSame"
    matcher = Diff(expected)
    cond, obj, message = matcher(orig)
    assert cond
    assert matcher.expected == expected
    assert obj == orig


def test_matchers_not_same_diff():
    expected = "Ahoj svet ako sa mas?\nJa sa mam zle.\nSame\n"
    orig = "Ahoj svet ako sa mas?\nJa sa mam dobre.\nNot same\n"
    matcher = Diff(expected)
    cond, obj, message = matcher(orig)
    assert not cond
    assert matcher.expected == expected
    assert obj == orig


def test_matchers_equals():
    matcher = Equals('expected string')
    cond, obj, message = matcher.invoke('expected string')
    assert cond
    assert matcher.expected == "expected string"
    assert obj == "expected string"


def test_matchers_not_equals():
    matcher = NotEquals('expected string')
    cond, obj, message = matcher.invoke('expected string')
    assert not cond
    assert matcher.expected == "expected string"
    assert obj == "expected string"


def test_matchers_greater():
    matcher = Greater(2)
    cond, obj, message = matcher(1)
    assert not cond


def test_matchers_less():
    matcher = Less(2)
    cond, obj, message = matcher(1)
    assert cond


def test_matchers_less_equals():
    matcher = LessEquals(2)
    cond, obj, message = matcher(2)
    cond2, obj2, message2 = matcher(1)
    assert cond2
    assert cond


def test_matchers_greater_equals():
    matcher = GreaterEquals(2)
    cond, obj, message = matcher(2)
    cond2, obj2, message2 = matcher(3)
    assert cond2
    assert cond


def test_matchers_is_none():
    matcher = IsNone()
    cond, obj, message = matcher(None)
    assert cond
    cond, obj, message = matcher(1111)
    assert not cond


def test_matchers_is_not_none():
    matcher = IsNotNone()
    cond, obj, message = matcher(None)
    assert not cond
    cond, obj, message = matcher(1111)
    assert cond


def test_matchers_checks_passed(task):
    matcher = ChecksPassed()
    task.asserts.check(True)
    cond, obj, message = matcher(task)
    assert cond


def test_matchers_checks_failed(task):
    matcher = ChecksFailed()
    task.asserts.check(False)
    cond, obj, message = matcher(task)
    assert cond


def test_matchers_result_passed():
    matcher = ResultPassed()
    passed_result = results.PASS
    cond, obj, message = matcher(passed_result)
    assert cond


def test_matchers_result_failed():
    matcher = ResultFailed()
    cond, obj, message = matcher(results.FAIL)
    assert cond



from ktdk.asserts.matchers import IsNotNone
from ktdk.core.tasks import Task
from ktdk.runtime.context import Context
from ktdk.runtime.runners import TaskRunner


class SimpleCounterLocalTask(Task):
    def _run(self, *args, **kwargs):
        self.context.config.set_test('lc', 1)
        self.asserts.check(self.context.config, IsNotNone())


class SimpleCounterGlobalTask(Task):
    def _run(self, *args, **kwargs):
        gc = self.context.config['gc']
        self.context.config.set_suite('gc', gc + 1)
        self.asserts.check(self.context.config, IsNotNone())


class FailingTask(Task):
    def _run(self, *args, **kwargs):
        self.asserts.check(None, IsNotNone())


class FailingRequiredTask(Task):
    def _run(self, *args, **kwargs):
        self.asserts.check(None, IsNotNone())


class FailingRequireTask(Task):
    def _run(self, *args, **kwargs):
        self.asserts.require(None, IsNotNone())


class ExceptionThrowingTask(Task):
    def _run(self, *args, **kwargs):
        raise RuntimeError("Exception")


def test_task_runner_with_one_task():
    task = SimpleCounterLocalTask('counter_task')
    context = Context({'gc': 0}, {'lc': 0})
    runner = TaskRunner(task=task, config={}, context=context)
    runner.invoke()
    assert context.config['lc'] == 1
    assert context.config['gc'] == 0
    assert task.result.effective.passed
    assert task.result.current.passed


def test_task_runner_with_two_tasks():
    task = SimpleCounterLocalTask('counter_local')
    gtask = SimpleCounterGlobalTask('counter_global')
    task.add_task(gtask)
    context = Context({'gc': 0}, {'lc': 0})
    runner = TaskRunner(task=task, config={}, context=context)
    runner.invoke()
    assert task.result.effective.passed
    assert task.result.current.passed
    assert context.config['lc'] == 1
    assert context.config['gc'] == 1
    assert task.asserts.passed


def test_task_runner_with_failing_task():
    task = FailingTask('failing_task')
    context = Context({'gc': 0}, {'lc': 0})
    runner = TaskRunner(task=task, config={}, context=context)
    runner.invoke()
    assert task.result.effective.failed
    assert task.asserts.failed


def test_task_runner_with_tasks_one_failed():
    task = SimpleCounterLocalTask('counter_local')
    gtask = FailingTask('failing_task')
    task.add_task(gtask)
    context = Context({'gc': 0}, {'lc': 0})
    runner = TaskRunner(task=task, config={}, context=context)
    runner.invoke()
    assert gtask.result.effective.failed
    assert gtask.result.current.failed
    assert task.result.current.passed
    assert task.result.effective.passed


def test_task_runner_with_tasks_failed_require():
    task = SimpleCounterLocalTask('counter_local')
    gtask = FailingRequireTask('failing_require_task')
    task.require_that(gtask)
    context = Context({'gc': 0}, {'lc': 0})
    runner = TaskRunner(task=task, config={}, context=context)
    runner.invoke()
    assert gtask.result.effective.failed
    assert gtask.result.current.failed
    assert task.result.current.passed
    assert task.result.effective.failed


def test_task_runner_with_exception_task():
    task = ExceptionThrowingTask('exception_throwing')
    context = Context({'gc': 0, 'devel': False}, {'lc': 0})
    runner = TaskRunner(task=task, config={}, context=context)
    runner.invoke()
    assert task.result.effective.errored
    assert task.result.current.errored



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

requirements = ['requests', 'pyjwt']
requirements = ['unidecode', 'junitparser', 'pyyaml', 'click', 'requests', 'coloredlogs']
extra_requirements = {
    'dev': ['pytest>=3', 'coverage', 'pytest-cov', 'pytest-mock', ], 'docs': ['sphinx', ]
}
entry_points = dict(console_scripts=['ktdk = ktdk.cli:main_cli', ])
setup(name='ktdk',
      version='0.1',
      description='Kontr tests development kit',
      author='Peter Stanko',
      author_email='stanko@mail.muni.cz',
      url='https://gitlab.fi.muni.cz/grp-kontr2/ktdk',
      packages=find_packages(exclude=("tests",)),
      long_description=long_description,
      long_description_content_type='text/markdown',
      include_package_data=True,
      install_requires=requirements,
      extras_require=extra_requirements,
      entry_points=entry_points,
      classifiers=[
          "Programming Language :: Python :: 3",
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          "Operating System :: OS Independent",
          "License :: OSI Approved :: Apache Software License",
          'Intended Audience :: Developers',
          'Topic :: Utilities',
      ],)



import logging

log = logging.getLogger(__name__)


class TagsEvaluator:
    def __init__(self, expression: str, registered_tags=None):
        self.expression = expression
        self.registered_tags = registered_tags or []

    def register_tags(self, *tags):
        self.registered_tags.extend(tags)

    def build_dict(self, *tags):
        result = {}
        for tag in self.registered_tags:
            result[tag] = tag in tags
        return result

    def evaluate(self, *test_tags):
        expression = self.expression
        if not expression:
            return True
        tags = self.build_dict(*test_tags)
        log.trace(f"[TAGS] Eval: \"{expression}\": {tags}")
        try:
            return eval(expression, {'__builtins__': None}, tags)
        except Exception as ex:
            log.warn(f"There is some error: {ex}")
            return False
